# Copyright 2020 The Magenta Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for polyphony_lib."""

import copy

from magenta.common import testing_lib as common_testing_lib
from magenta.models.polyphony_rnn import polyphony_lib
from note_seq import sequences_lib
from note_seq import testing_lib
from note_seq.protobuf import music_pb2
import tensorflow.compat.v1 as tf

tf.disable_v2_behavior()


class PolyphonyLibTest(tf.test.TestCase):

  def setUp(self):
    self.maxDiff = None  # pylint:disable=invalid-name

    self.note_sequence = common_testing_lib.parse_test_proto(
        music_pb2.NoteSequence,
        """
        tempos: {
          qpm: 60
        }
        ticks_per_quarter: 220
        """)

  def testFromQuantizedNoteSequence(self):
    testing_lib.add_track_to_sequence(
        self.note_sequence, 0,
        [(60, 100, 0.0, 4.0), (64, 100, 0.0, 3.0), (67, 100, 1.0, 2.0)])
    quantized_sequence = sequences_lib.quantize_note_sequence(
        self.note_sequence, steps_per_quarter=1)
    poly_seq = list(polyphony_lib.PolyphonicSequence(quantized_sequence))

    pe = polyphony_lib.PolyphonicEvent
    expected_poly_seq = [
        pe(pe.START, None),
        # step 0
        pe(pe.NEW_NOTE, 64),
        pe(pe.NEW_NOTE, 60),
        pe(pe.STEP_END, None),
        # step 1
        pe(pe.NEW_NOTE, 67),
        pe(pe.CONTINUED_NOTE, 64),
        pe(pe.CONTINUED_NOTE, 60),
        pe(pe.STEP_END, None),
        # step 2
        pe(pe.CONTINUED_NOTE, 64),
        pe(pe.CONTINUED_NOTE, 60),
        pe(pe.STEP_END, None),
        # step 3
        pe(pe.CONTINUED_NOTE, 60),
        pe(pe.STEP_END, None),

        pe(pe.END, None),
    ]
    self.assertEqual(expected_poly_seq, poly_seq)

  def testToSequence(self):
    testing_lib.add_track_to_sequence(
        self.note_sequence, 0,
        [(60, 100, 0.0, 4.0), (64, 100, 0.0, 3.0), (67, 100, 1.0, 2.0)])
    quantized_sequence = sequences_lib.quantize_note_sequence(
        self.note_sequence, steps_per_quarter=1)
    poly_seq = polyphony_lib.PolyphonicSequence(quantized_sequence)
    poly_seq_ns = poly_seq.to_sequence(qpm=60.0)

    # Make comparison easier
    poly_seq_ns.notes.sort(key=lambda n: (n.start_time, n.pitch))
    self.note_sequence.notes.sort(key=lambda n: (n.start_time, n.pitch))

    self.assertEqual(self.note_sequence, poly_seq_ns)

  def testToSequenceWithContinuedNotesNotStarted(self):
    poly_seq = polyphony_lib.PolyphonicSequence(steps_per_quarter=1)

    pe = polyphony_lib.PolyphonicEvent
    poly_events = [
        # step 0
        pe(pe.NEW_NOTE, 60),
        pe(pe.NEW_NOTE, 64),
        pe(pe.STEP_END, None),
        # step 1
        pe(pe.CONTINUED_NOTE, 60),
        pe(pe.CONTINUED_NOTE, 64),
        pe(pe.CONTINUED_NOTE, 67),  # Was not started, should be ignored.
        pe(pe.STEP_END, None),

        pe(pe.END, None),
    ]
    for event in poly_events:
      poly_seq.append(event)

    poly_seq_ns = poly_seq.to_sequence(qpm=60.0)

    testing_lib.add_track_to_sequence(
        self.note_sequence, 0,
        [(60, 100, 0.0, 2.0), (64, 100, 0.0, 2.0)])

    # Make comparison easier
    poly_seq_ns.notes.sort(key=lambda n: (n.start_time, n.pitch))
    self.note_sequence.notes.sort(key=lambda n: (n.start_time, n.pitch))

    self.assertEqual(self.note_sequence, poly_seq_ns)

  def testToSequenceWithExtraEndEvents(self):
    poly_seq = polyphony_lib.PolyphonicSequence(steps_per_quarter=1)

    pe = polyphony_lib.PolyphonicEvent
    poly_events = [
        # step 0
        pe(pe.NEW_NOTE, 60),
        pe(pe.END, None),  # END event before end. Should be ignored.
        pe(pe.NEW_NOTE, 64),
        pe(pe.END, None),  # END event before end. Should be ignored.
        pe(pe.STEP_END, None),
        pe(pe.END, None),  # END event before end. Should be ignored.
        # step 1
        pe(pe.CONTINUED_NOTE, 60),
        pe(pe.END, None),  # END event before end. Should be ignored.
        pe(pe.CONTINUED_NOTE, 64),
        pe(pe.END, None),  # END event before end. Should be ignored.
        pe(pe.STEP_END, None),

        pe(pe.END, None),
    ]
    for event in poly_events:
      poly_seq.append(event)

    poly_seq_ns = poly_seq.to_sequence(qpm=60.0)

    testing_lib.add_track_to_sequence(
        self.note_sequence, 0,
        [(60, 100, 0.0, 2.0), (64, 100, 0.0, 2.0)])

    # Make comparison easier
    poly_seq_ns.notes.sort(key=lambda n: (n.start_time, n.pitch))
    self.note_sequence.notes.sort(key=lambda n: (n.start_time, n.pitch))

    self.assertEqual(self.note_sequence, poly_seq_ns)

  def testToSequenceWithUnfinishedSequence(self):
    poly_seq = polyphony_lib.PolyphonicSequence(steps_per_quarter=1)

    pe = polyphony_lib.PolyphonicEvent
    poly_events = [
        # step 0
        pe(pe.NEW_NOTE, 60),
        pe(pe.NEW_NOTE, 64),
        # missing STEP_END and END events at end of sequence.
    ]
    for event in poly_events:
      poly_seq.append(event)

    with self.assertRaises(ValueError):
      poly_seq.to_sequence(qpm=60.0)

  def testToSequenceWithRepeatedNotes(self):
    poly_seq = polyphony_lib.PolyphonicSequence(steps_per_quarter=1)

    pe = polyphony_lib.PolyphonicEvent
    poly_events = [
        # step 0
        pe(pe.NEW_NOTE, 60),
        pe(pe.NEW_NOTE, 64),
        pe(pe.STEP_END, None),
        # step 1
        pe(pe.NEW_NOTE, 60),
        pe(pe.CONTINUED_NOTE, 64),
        pe(pe.STEP_END, None),

        pe(pe.END, None),
    ]
    for event in poly_events:
      poly_seq.append(event)

    poly_seq_ns = poly_seq.to_sequence(qpm=60.0)

    testing_lib.add_track_to_sequence(
        self.note_sequence, 0,
        [(60, 100, 0.0, 1.0), (64, 100, 0.0, 2.0), (60, 100, 1.0, 2.0)])

    # Make comparison easier
    poly_seq_ns.notes.sort(key=lambda n: (n.start_time, n.pitch))
    self.note_sequence.notes.sort(key=lambda n: (n.start_time, n.pitch))

    self.assertEqual(self.note_sequence, poly_seq_ns)

  def testToSequenceWithBaseNoteSequence(self):
    poly_seq = polyphony_lib.PolyphonicSequence(
        steps_per_quarter=1, start_step=1)

    pe = polyphony_lib.PolyphonicEvent
    poly_events = [
        # step 0
        pe(pe.NEW_NOTE, 60),
        pe(pe.NEW_NOTE, 64),
        pe(pe.STEP_END, None),
        # step 1
        pe(pe.CONTINUED_NOTE, 60),
        pe(pe.CONTINUED_NOTE, 64),
        pe(pe.STEP_END, None),

        pe(pe.END, None),
    ]
    for event in poly_events:
      poly_seq.append(event)

    base_seq = copy.deepcopy(self.note_sequence)
    testing_lib.add_track_to_sequence(
        base_seq, 0, [(60, 100, 0.0, 1.0)])

    poly_seq_ns = poly_seq.to_sequence(qpm=60.0, base_note_sequence=base_seq)

    testing_lib.add_track_to_sequence(
        self.note_sequence, 0,
        [(60, 100, 0.0, 1.0), (60, 100, 1.0, 3.0), (64, 100, 1.0, 3.0)])

    # Make comparison easier
    poly_seq_ns.notes.sort(key=lambda n: (n.start_time, n.pitch))
    self.note_sequence.notes.sort(key=lambda n: (n.start_time, n.pitch))

    self.assertEqual(self.note_sequence, poly_seq_ns)

  def testToSequenceWithEmptySteps(self):
    poly_seq = polyphony_lib.PolyphonicSequence(
        steps_per_quarter=1)

    pe = polyphony_lib.PolyphonicEvent
    poly_events = [
        # step 0
        pe(pe.STEP_END, None),
        # step 1
        pe(pe.STEP_END, None),

        pe(pe.END, None),
    ]
    for event in poly_events:
      poly_seq.append(event)

    poly_seq_ns = poly_seq.to_sequence(qpm=60.0)

    self.note_sequence.total_time = 2

    self.assertEqual(self.note_sequence, poly_seq_ns)

  def testSetLengthAddSteps(self):
    poly_seq = polyphony_lib.PolyphonicSequence(steps_per_quarter=1)
    poly_seq.set_length(5)

    self.assertEqual(5, poly_seq.num_steps)
    self.assertListEqual([0, 0, 1, 2, 3, 4, 5], poly_seq.steps)

    pe = polyphony_lib.PolyphonicEvent
    poly_events = [
        pe(pe.START, None),

        pe(pe.STEP_END, None),
        pe(pe.STEP_END, None),
        pe(pe.STEP_END, None),
        pe(pe.STEP_END, None),
        pe(pe.STEP_END, None),

        pe(pe.END, None),
    ]
    self.assertEqual(poly_events, list(poly_seq))

    # Add 5 more steps to make sure END is managed properly.
    poly_seq.set_length(10)

    self.assertEqual(10, poly_seq.num_steps)
    self.assertListEqual([0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10], poly_seq.steps)

    pe = polyphony_lib.PolyphonicEvent
    poly_events = [
        pe(pe.START, None),

        pe(pe.STEP_END, None),
        pe(pe.STEP_END, None),
        pe(pe.STEP_END, None),
        pe(pe.STEP_END, None),
        pe(pe.STEP_END, None),
        pe(pe.STEP_END, None),
        pe(pe.STEP_END, None),
        pe(pe.STEP_END, None),
        pe(pe.STEP_END, None),
        pe(pe.STEP_END, None),

        pe(pe.END, None),
    ]
    self.assertEqual(poly_events, list(poly_seq))

  def testSetLengthAddStepsToSequenceWithoutEnd(self):
    poly_seq = polyphony_lib.PolyphonicSequence(steps_per_quarter=1)

    # Construct a list with one silence step and no END.
    pe = polyphony_lib.PolyphonicEvent
    poly_seq.append(pe(pe.STEP_END, None))

    poly_seq.set_length(2)
    poly_events = [
        pe(pe.START, None),

        pe(pe.STEP_END, None),
        pe(pe.STEP_END, None),

        pe(pe.END, None),
    ]
    self.assertEqual(poly_events, list(poly_seq))

  def testSetLengthAddStepsToSequenceWithUnfinishedStep(self):
    poly_seq = polyphony_lib.PolyphonicSequence(steps_per_quarter=1)

    # Construct a list with one note and no STEP_END or END.
    pe = polyphony_lib.PolyphonicEvent
    poly_seq.append(pe(pe.NEW_NOTE, 60))

    poly_seq.set_length(2)
    poly_events = [
        pe(pe.START, None),

        pe(pe.NEW_NOTE, 60),
        pe(pe.STEP_END, None),

        pe(pe.STEP_END, None),

        pe(pe.END, None),
    ]
    self.assertEqual(poly_events, list(poly_seq))

  def testSetLengthRemoveSteps(self):
    poly_seq = polyphony_lib.PolyphonicSequence(steps_per_quarter=1)

    pe = polyphony_lib.PolyphonicEvent
    poly_events = [
        # step 0
        pe(pe.NEW_NOTE, 60),
        pe(pe.STEP_END, None),
        # step 1
        pe(pe.NEW_NOTE, 64),
        pe(pe.STEP_END, None),
        # step 2
        pe(pe.NEW_NOTE, 67),
        pe(pe.STEP_END, None),

        pe(pe.END, None),
    ]
    for event in poly_events:
      poly_seq.append(event)

    poly_seq.set_length(2)
    poly_events = [
        pe(pe.START, None),
        # step 0
        pe(pe.NEW_NOTE, 60),
        pe(pe.STEP_END, None),
        # step 1
        pe(pe.NEW_NOTE, 64),
        pe(pe.STEP_END, None),

        pe(pe.END, None),
    ]
    self.assertEqual(poly_events, list(poly_seq))

    poly_seq.set_length(1)
    poly_events = [
        pe(pe.START, None),
        # step 0
        pe(pe.NEW_NOTE, 60),
        pe(pe.STEP_END, None),

        pe(pe.END, None),
    ]
    self.assertEqual(poly_events, list(poly_seq))

    poly_seq.set_length(0)
    poly_events = [
        pe(pe.START, None),

        pe(pe.END, None),
    ]
    self.assertEqual(poly_events, list(poly_seq))

  def testSetLengthRemoveStepsFromSequenceWithoutEnd(self):
    poly_seq = polyphony_lib.PolyphonicSequence(steps_per_quarter=1)

    # Construct a list with two silence steps and no END.
    pe = polyphony_lib.PolyphonicEvent
    poly_seq.append(pe(pe.STEP_END, None))
    poly_seq.append(pe(pe.STEP_END, None))

    poly_seq.set_length(1)
    poly_events = [
        pe(pe.START, None),

        pe(pe.STEP_END, None),

        pe(pe.END, None),
    ]
    self.assertEqual(poly_events, list(poly_seq))

  def testSetLengthRemoveStepsFromSequenceWithUnfinishedStep(self):
    poly_seq = polyphony_lib.PolyphonicSequence(steps_per_quarter=1)

    # Construct a list with a silence step, a new note, and no STEP_END or END.
    pe = polyphony_lib.PolyphonicEvent
    poly_seq.append(pe(pe.STEP_END, None))
    poly_seq.append(pe(pe.NEW_NOTE, 60))

    poly_seq.set_length(1)
    poly_events = [
        pe(pe.START, None),

        pe(pe.STEP_END, None),

        pe(pe.END, None),
    ]
    self.assertEqual(poly_events, list(poly_seq))

  def testNumSteps(self):
    poly_seq = polyphony_lib.PolyphonicSequence(steps_per_quarter=1)

    pe = polyphony_lib.PolyphonicEvent
    poly_events = [
        # step 0
        pe(pe.NEW_NOTE, 60),
        pe(pe.NEW_NOTE, 64),
        pe(pe.STEP_END, None),
        # step 1
        pe(pe.CONTINUED_NOTE, 60),
        pe(pe.CONTINUED_NOTE, 64),
        pe(pe.STEP_END, None),

        pe(pe.END, None),
    ]
    for event in poly_events:
      poly_seq.append(event)

    self.assertEqual(2, poly_seq.num_steps)
    self.assertListEqual([0, 0, 0, 0, 1, 1, 1, 2], poly_seq.steps)

  def testNumStepsIncompleteStep(self):
    poly_seq = polyphony_lib.PolyphonicSequence(steps_per_quarter=1)

    pe = polyphony_lib.PolyphonicEvent
    poly_events = [
        # step 0
        pe(pe.NEW_NOTE, 60),
        pe(pe.NEW_NOTE, 64),
        pe(pe.STEP_END, None),
        # step 1
        pe(pe.CONTINUED_NOTE, 60),
        pe(pe.CONTINUED_NOTE, 64),
        pe(pe.STEP_END, None),
        # incomplete step. should not be counted.
        pe(pe.NEW_NOTE, 72),

    ]
    for event in poly_events:
      poly_seq.append(event)

    self.assertEqual(2, poly_seq.num_steps)
    self.assertListEqual([0, 0, 0, 0, 1, 1, 1, 2], poly_seq.steps)

  def testSteps(self):
    pe = polyphony_lib.PolyphonicEvent
    poly_events = [
        # step 0
        pe(pe.NEW_NOTE, 60),
        pe(pe.NEW_NOTE, 64),
        pe(pe.STEP_END, None),
        # step 1
        pe(pe.CONTINUED_NOTE, 60),
        pe(pe.CONTINUED_NOTE, 64),
        pe(pe.STEP_END, None),

        pe(pe.END, None),
    ]

    poly_seq = polyphony_lib.PolyphonicSequence(steps_per_quarter=1)
    for event in poly_events:
      poly_seq.append(event)
    self.assertListEqual([0, 0, 0, 0, 1, 1, 1, 2], poly_seq.steps)

    poly_seq = polyphony_lib.PolyphonicSequence(
        steps_per_quarter=1, start_step=2)
    for event in poly_events:
      poly_seq.append(event)
    self.assertListEqual([2, 2, 2, 2, 3, 3, 3, 4], poly_seq.steps)

  def testTrimTrailingEndEvents(self):
    poly_seq = polyphony_lib.PolyphonicSequence(steps_per_quarter=1)

    pe = polyphony_lib.PolyphonicEvent
    poly_events = [
        # step 0
        pe(pe.NEW_NOTE, 60),
        pe(pe.STEP_END, None),

        pe(pe.END, None),
        pe(pe.END, None),
    ]
    for event in poly_events:
      poly_seq.append(event)

    poly_seq.trim_trailing_end_events()

    poly_events_expected = [
        pe(pe.START, None),
        # step 0
        pe(pe.NEW_NOTE, 60),
        pe(pe.STEP_END, None),
    ]

    self.assertEqual(poly_events_expected, list(poly_seq))

  def testExtractPolyphonicSequences(self):
    testing_lib.add_track_to_sequence(
        self.note_sequence, 0, [(60, 100, 0.0, 4.0)])
    quantized_sequence = sequences_lib.quantize_note_sequence(
        self.note_sequence, steps_per_quarter=1)

    seqs, _ = polyphony_lib.extract_polyphonic_sequences(quantized_sequence)
    self.assertEqual(1, len(seqs))

    seqs, _ = polyphony_lib.extract_polyphonic_sequences(
        quantized_sequence, min_steps_discard=2, max_steps_discard=5)
    self.assertEqual(1, len(seqs))

    self.note_sequence.notes[0].end_time = 1.0
    self.note_sequence.total_time = 1.0
    quantized_sequence = sequences_lib.quantize_note_sequence(
        self.note_sequence, steps_per_quarter=1)
    seqs, _ = polyphony_lib.extract_polyphonic_sequences(
        quantized_sequence, min_steps_discard=3, max_steps_discard=5)
    self.assertEqual(0, len(seqs))

    self.note_sequence.notes[0].end_time = 10.0
    self.note_sequence.total_time = 10.0
    quantized_sequence = sequences_lib.quantize_note_sequence(
        self.note_sequence, steps_per_quarter=1)
    seqs, _ = polyphony_lib.extract_polyphonic_sequences(
        quantized_sequence, min_steps_discard=3, max_steps_discard=5)
    self.assertEqual(0, len(seqs))

  def testExtractPolyphonicMultiProgram(self):
    testing_lib.add_track_to_sequence(
        self.note_sequence, 0,
        [(60, 100, 0.0, 4.0), (64, 100, 0.0, 3.0), (67, 100, 1.0, 2.0)])
    self.note_sequence.notes[0].program = 2
    quantized_sequence = sequences_lib.quantize_note_sequence(
        self.note_sequence, steps_per_quarter=1)

    seqs, _ = polyphony_lib.extract_polyphonic_sequences(quantized_sequence)
    self.assertEqual(0, len(seqs))

  def testExtractNonZeroStart(self):
    testing_lib.add_track_to_sequence(
        self.note_sequence, 0, [(60, 100, 0.0, 4.0)])
    quantized_sequence = sequences_lib.quantize_note_sequence(
        self.note_sequence, steps_per_quarter=1)

    seqs, _ = polyphony_lib.extract_polyphonic_sequences(
        quantized_sequence, start_step=4, min_steps_discard=1)
    self.assertEqual(0, len(seqs))
    seqs, _ = polyphony_lib.extract_polyphonic_sequences(
        quantized_sequence, start_step=0, min_steps_discard=1)
    self.assertEqual(1, len(seqs))


if __name__ == '__main__':
  tf.test.main()



# Copyright 2020 The Magenta Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Defines the Material Design Icons Problem."""
import io
import numpy as np

from PIL import Image
from six.moves import zip_longest
from skimage import draw

import tensorflow.compat.v1 as tf


SVG_PREFIX_BIG = ('<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="'
                  'http://www.w3.org/1999/xlink" width="256px" height="256px"'
                  ' style="-ms-transform: rotate(360deg); -webkit-transform:'
                  ' rotate(360deg); transform: rotate(360deg);" '
                  'preserveAspectRatio="xMidYMid meet" viewBox="0 0 24 24">')
PATH_PREFIX_1 = '<path d="'
PATH_POSFIX_1 = '" fill="currentColor"/>'
SVG_POSFIX = '</svg>'

NUM_ARGS = {'v': 1, 'V': 1, 'h': 1, 'H': 1, 'a': 7, 'A': 7, 'l': 2, 'L': 2,
            't': 2, 'T': 2, 'c': 6, 'C': 6, 'm': 2, 'M': 2, 's': 4, 'S': 4,
            'q': 4, 'Q': 4, 'z': 0}
# in order of arg complexity, with absolutes clustered
# recall we don't handle all commands (see docstring)
CMDS_LIST = 'zhvmltsqcaHVMLTSQCA'
CMD_MAPPING = {cmd: i for i, cmd in enumerate(CMDS_LIST)}


############################### GENERAL UTILS #################################
def grouper(iterable, batch_size, fill_value=None):
  """Helper method for returning batches of size batch_size of a dataset."""
  # grouper('ABCDEF', 3) -> 'ABC', 'DEF'
  args = [iter(iterable)] * batch_size
  return zip_longest(*args, fillvalue=fill_value)


def map_uni_to_alphanum(uni):
  """Maps [0-9 A-Z a-z] to numbers 0-62."""
  if 48 <= uni <= 57:
    return uni - 48
  elif 65 <= uni <= 90:
    return uni - 65 + 10
  return uni - 97 + 36


############# UTILS FOR CONVERTING SFD/SPLINESETS TO SVG PATHS ################
def _get_spline(sfd):
  if 'SplineSet' not in sfd:
    return ''
  pro = sfd[sfd.index('SplineSet')+10:]
  pro = pro[:pro.index('EndSplineSet')]
  return pro


def _spline_to_path_list(spline, height, replace_with_prev=False):
  """Converts SplineSet to a list of tokenized commands in svg path."""
  path = []
  prev_xy = []
  for line in spline.splitlines():
    if not line:
      continue
    tokens = line.split(' ')
    cmd = tokens[-2]
    if cmd not in 'cml':
      # COMMAND NOT RECOGNIZED.
      return []
      # assert cmd in 'cml', 'Command not recognized: {}'.format(cmd)
    args = tokens[:-2]
    args = [float(x) for x in args if x]

    if replace_with_prev and cmd in 'c':
      args[:2] = prev_xy
    prev_xy = args[-2:]

    new_y_args = []
    for i, a in enumerate(args):
      if i % 2 == 1:
        new_y_args.append((height - a))
      else:
        new_y_args.append((a))

    path.append([cmd.upper()] + new_y_args)
  return path


def sfd_to_path_list(single, replace_with_prev=False):
  """Converts the given SFD glyph into a path."""
  return _spline_to_path_list(_get_spline(single['sfd']),
                              single['vwidth'], replace_with_prev)


#################### UTILS FOR PROCESSING TOKENIZED PATHS #####################
def add_missing_cmds(path, remove_zs=False):
  """Adds missing cmd tags to the commands in the svg."""
  # For instance, the command 'a' takes 7 arguments, but some SVGs declare:
  #   a 1 2 3 4 5 6 7 8 9 10 11 12 13 14
  # Which is 14 arguments. This function converts the above to the equivalent:
  #   a 1 2 3 4 5 6 7  a 8 9 10 11 12 13 14
  #
  # Note: if remove_zs is True, this also removes any occurences of z commands.
  new_path = []
  for cmd in path:
    if not remove_zs or cmd[0] not in 'Zz':
      for new_cmd in add_missing_cmd(cmd):
        new_path.append(new_cmd)
  return new_path


def add_missing_cmd(command_list):
  """Adds missing cmd tags to the given command list."""
  # E.g.: given:
  #   ['a', '0', '0', '0', '0', '0', '0', '0',
  #         '0', '0', '0', '0', '0', '0', '0']
  # Converts to:
  #   [['a', '0', '0', '0', '0', '0', '0', '0'],
  #    ['a', '0', '0', '0', '0', '0', '0', '0']]
  # And returns a string that joins these elements with spaces.
  cmd_tag = command_list[0]
  args = command_list[1:]

  final_cmds = []
  for arg_batch in grouper(args, NUM_ARGS[cmd_tag]):
    final_cmds.append([cmd_tag] + list(arg_batch))

  if not final_cmds:
    # command has no args (e.g.: 'z')
    final_cmds = [[cmd_tag]]

  return final_cmds


def _normalize_args(arglist, norm, add=None, flip=False):
  """Normalize the given args with the given norm value."""
  new_arglist = []
  for i, arg in enumerate(arglist):
    new_arg = float(arg)

    if add is not None:
      add_to_x, add_to_y = add

      # This argument is an x-coordinate if even, y-coordinate if odd
      # except when flip == True
      if i % 2 == 0:
        new_arg += add_to_y if flip else add_to_x
      else:
        new_arg += add_to_x if flip else add_to_y

    new_arglist.append(str(24 * new_arg / norm))
  return new_arglist


def normalize_based_on_viewbox(path, viewbox):
  """Normalizes all args in a path to a standard 24x24 viewbox."""
  # Each SVG lives in a 2D plane. The viewbox determines the region of that
  # plane that gets rendered. For instance, some designers may work with a
  # viewbox that's 24x24, others with one that's 100x100, etc.

  # Suppose I design the the letter "h" in the Arial style using a 100x100
  # viewbox (let's call it icon A). Let's suppose the icon has height 75. Then,
  # I design the same character using a 20x20 viewbox (call this icon B), with
  # height 15 (=75% of 20). This means that, when rendered, both icons with look
  # exactly the same, but the scale of the commands each icon is using is
  # different. For instance, if icon A has a command like "lineTo 100 100", the
  # equivalent command in icon B will be "lineTo 20 20".

  # In order to avoid this problem and bring all real values to the same scale,
  # I scale all icons' commands to use a 24x24 viewbox. This function does this:
  # it converts a path that exists in the given viewbox into a standard 24x24
  # viewbox.
  viewbox = viewbox.split(' ')
  norm = max(int(viewbox[-1]), int(viewbox[-2]))

  if int(viewbox[-1]) > int(viewbox[-2]):
    add_to_y = 0
    add_to_x = abs(int(viewbox[-1]) - int(viewbox[-2])) / 2
  else:
    add_to_y = abs(int(viewbox[-1]) - int(viewbox[-2])) / 2
    add_to_x = 0

  new_path = []
  for command in path:
    if command[0] == 'a':
      new_path.append([command[0]] +
                      _normalize_args(command[1:3], norm) +
                      command[3:6] +
                      _normalize_args(command[6:], norm))
    elif command[0] == 'A':
      new_path.append(
          [command[0]] +
          _normalize_args(command[1:3], norm) +
          command[3:6] +
          _normalize_args(command[6:], norm, add=(add_to_x, add_to_y)))
    elif command[0] == 'V':
      new_path.append(
          [command[0]] +
          _normalize_args(command[1:], norm, add=(add_to_x, add_to_y),
                          flip=True))
    elif command[0] == command[0].upper():
      new_path.append(
          [command[0]] +
          _normalize_args(command[1:], norm, add=(add_to_x, add_to_y)))
    elif command[0] in 'zZ':
      new_path.append([command[0]])
    else:
      new_path.append([command[0]] + _normalize_args(command[1:], norm))

  return new_path


def _convert_args(args, curr_pos, cmd):
  """Converts given args to relative values."""
  # NOTE: glyphs only use a very small subset of commands (L, C, M, and Z -- I
  # believe). So I'm not handling A and H for now.
  if cmd in 'AH':
    raise NotImplementedError('These commands have >6 args (not supported).')

  new_args = []
  for i, arg in enumerate(args):
    x_or_y = i % 2
    if cmd == 'H':
      x_or_y = (i + 1) % 2
    new_args.append(str(float(arg) - curr_pos[x_or_y]))

  return new_args


def _update_curr_pos(curr_pos, cmd, start_of_path):
  """Calculate the position of the pen after cmd is applied."""
  if cmd[0] in 'ml':
    curr_pos = [curr_pos[0] + float(cmd[1]), curr_pos[1] + float(cmd[2])]
    if cmd[0] == 'm':
      start_of_path = curr_pos
  elif cmd[0] in 'z':
    curr_pos = start_of_path
  elif cmd[0] in 'h':
    curr_pos = [curr_pos[0] + float(cmd[1]), curr_pos[1]]
  elif cmd[0] in 'v':
    curr_pos = [curr_pos[0], curr_pos[1] + float(cmd[1])]
  elif cmd[0] in 'ctsqa':
    curr_pos = [curr_pos[0] + float(cmd[-2]), curr_pos[1] + float(cmd[-1])]

  return curr_pos, start_of_path


def make_relative(cmds):
  """Convert commands in a path to relative positioning."""
  curr_pos = (0.0, 0.0)
  start_of_path = (0.0, 0.0)
  new_cmds = []
  for cmd in cmds:
    if cmd[0].lower() == cmd[0]:
      new_cmd = cmd
    elif cmd[0].lower() == 'z':
      new_cmd = [cmd[0].lower()]
    else:
      new_cmd = [cmd[0].lower()] + _convert_args(cmd[1:], curr_pos, cmd=cmd[0])
    new_cmds.append(new_cmd)
    curr_pos, start_of_path = _update_curr_pos(curr_pos, new_cmd, start_of_path)
  return new_cmds


def _is_to_left_of(pt1, pt2):
  pt1_norm = (pt1[0]**2 + pt1[1]**2)
  pt2_norm = (pt2[0]**2 + pt2[1]**2)
  return pt1[1] < pt2[1] or (pt1_norm == pt2_norm and pt1[0] < pt2[0])


def _get_leftmost_point(path):
  """Returns the leftmost, topmost point of the path."""
  leftmost = (float('inf'), float('inf'))
  idx = -1

  for i, cmd in enumerate(path):
    if len(cmd) > 1:
      endpoint = cmd[-2:]
      if _is_to_left_of(endpoint, leftmost):
        leftmost = endpoint
        idx = i

  return leftmost, idx


def _separate_substructures(path):
  """Returns a list of subpaths, each representing substructures the glyph."""
  substructures = []
  curr = []
  for cmd in path:
    if cmd[0] in 'mM' and curr:
      substructures.append(curr)
      curr = []
    curr.append(cmd)
  if curr:
    substructures.append(curr)
  return substructures


def _is_clockwise(subpath):
  """Returns whether the given subpath is clockwise-oriented."""
  pts = [cmd[-2:] for cmd in subpath]
  det = 0
  for i in range(len(pts)-1):
    det += np.linalg.det(pts[i:i+2])
  return det > 0


def _make_clockwise(subpath):
  """Inverts the cardinality of the given subpath."""
  new_path = [subpath[0]]
  other_cmds = list(reversed(subpath[1:]))
  for i, cmd in enumerate(other_cmds):
    if i+1 == len(other_cmds):
      where_we_were = subpath[0][-2:]
    else:
      where_we_were = other_cmds[i+1][-2:]

    if len(cmd) > 3:
      new_cmd = [cmd[0], cmd[3], cmd[4], cmd[1], cmd[2],
                 where_we_were[0], where_we_were[1]]
    else:
      new_cmd = [cmd[0], where_we_were[0], where_we_were[1]]

    new_path.append(new_cmd)
  return new_path


def canonicalize(path):
  """Makes all paths start at top left, and go clockwise first."""
  # convert args to floats
  path = [[x[0]] + list(map(float, x[1:])) for x in path]

  # canonicalize each subpath separately
  new_substructures = []
  for subpath in _separate_substructures(path):
    leftmost_point, leftmost_idx = _get_leftmost_point(subpath)
    reordered = ([['M', leftmost_point[0], leftmost_point[1]]] +
                 subpath[leftmost_idx+1:] + subpath[1:leftmost_idx+1])
    new_substructures.append((reordered, leftmost_point))

  new_path = []
  first_substructure_done = False
  should_flip_cardinality = False
  for sp, _ in sorted(new_substructures, key=lambda x: (x[1][1], x[1][0])):
    if not first_substructure_done:
      # we're looking at the first substructure now, we can determine whether we
      # will flip the cardniality of the whole icon or not
      should_flip_cardinality = not _is_clockwise(sp)
      first_substructure_done = True

    if should_flip_cardinality:
      sp = _make_clockwise(sp)

    new_path.extend(sp)

  # convert args to strs
  path = [[x[0]] + list(map(str, x[1:])) for x in new_path]
  return path


########## UTILS FOR CONVERTING TOKENIZED PATHS TO VECTORS ###########
def path_to_vector(path, categorical=False):
  """Converts path's commands to a series of vectors."""
  # Notes:
  #   - The SimpleSVG dataset does not have any 't', 'q', 'Z', 'T', or 'Q'.
  #     Thus, we don't handle those here.
  #   - We also removed all 'z's.
  #   - The x-axis-rotation argument to a commands is always 0 in this
  #     dataset, so we ignore it

  # Many commands have args that correspond to args in other commands.
  #   v  __,__ _______________ ______________,_________ __,__ __,__ _,y
  #   h  __,__ _______________ ______________,_________ __,__ __,__ x,_
  #   z  __,__ _______________ ______________,_________ __,__ __,__ _,_
  #   a  rx,ry x-axis-rotation large-arc-flag,sweepflag __,__ __,__ x,y
  #   l  __,__ _______________ ______________,_________ __,__ __,__ x,y
  #   c  __,__ _______________ ______________,_________ x1,y1 x2,y2 x,y
  #   m  __,__ _______________ ______________,_________ __,__ __,__ x,y
  #   s  __,__ _______________ ______________,_________ __,__ x2,y2 x,y

  # So each command will be converted to a vector where the dimension is the
  # minimal number of arguments to all commands:
  #   [rx, ry, large-arc-flag, sweepflag, x1, y1, x2, y2, x, y]
  # If a command does not output a certain arg, it is set to 0.
  #   "l 5,5" becomes [0, 0, 0, 0, 0, 0, 0, 0, 5, 5]

  # Also note, as of now we also output an extra dimension at index 0, which
  # indicates which command is being outputted (integer).
  new_path = []
  for cmd in path:
    new_path.append(cmd_to_vector(cmd, categorical=categorical))
  return new_path


def cmd_to_vector(cmd_list, categorical=False):
  """Converts the given command (given as a list) into a vector."""
  # For description of how this conversion happens, see
  # path_to_vector docstring.
  cmd = cmd_list[0]
  args = cmd_list[1:]

  if not categorical:
    # integer, for MSE
    command = [float(CMD_MAPPING[cmd])]
  else:
    # one hot + 1 dim for EOS.
    command = [0.0] * (len(CMDS_LIST) + 1)
    command[CMD_MAPPING[cmd] + 1] = 1.0

  arguments = [0.0] * 10
  if cmd in 'hH':
    arguments[8] = float(args[0])  # x
  elif cmd in 'vV':
    arguments[9] = float(args[0])  # y
  elif cmd in 'mMlLtT':
    arguments[8] = float(args[0])  # x
    arguments[9] = float(args[1])  # y
  elif cmd in 'sSqQ':
    arguments[6] = float(args[0])  # x2
    arguments[7] = float(args[1])  # y2
    arguments[8] = float(args[2])  # x
    arguments[9] = float(args[3])  # y
  elif cmd in 'cC':
    arguments[4] = float(args[0])  # x1
    arguments[5] = float(args[1])  # y1
    arguments[6] = float(args[2])  # x2
    arguments[7] = float(args[3])  # y2
    arguments[8] = float(args[4])  # x
    arguments[9] = float(args[5])  # y
  elif cmd in 'aA':
    arguments[0] = float(args[0])  # rx
    arguments[1] = float(args[1])  # ry
    # we skip x-axis-rotation
    arguments[2] = float(args[3])  # large-arc-flag
    arguments[3] = float(args[4])  # sweep-flag
    # a does not have x1, y1, x2, y2 args
    arguments[8] = float(args[5])  # x
    arguments[9] = float(args[6])  # y

  return command + arguments


################# UTILS FOR RENDERING PATH INTO IMAGE #################
def _cubicbezier(x0, y0, x1, y1, x2, y2, x3, y3, n=40):
  """Return n points along cubiz bezier with given control points."""
  # from http://rosettacode.org/wiki/Bitmap/B%C3%A9zier_curves/Cubic
  pts = []
  for i in range(n+1):
    t = float(i) / float(n)
    a = (1. - t)**3
    b = 3. * t * (1. - t)**2
    c = 3.0 * t**2 * (1.0 - t)
    d = t**3

    x = float(a * x0 + b * x1 + c * x2 + d * x3)
    y = float(a * y0 + b * y1 + c * y2 + d * y3)
    pts.append((x, y))
  return list(zip(*pts))


def _update_pos(curr_pos, end_pos, absolute):
  if absolute:
    return end_pos
  return curr_pos[0] + end_pos[0], curr_pos[1] + end_pos[1]


def constant_color(*unused_args):
  return np.array([255, 255, 255])


def _render_cubic(canvas, curr_pos, c_args, absolute, color):
  """Renders a cubic bezier curve in the given canvas."""
  if not absolute:
    c_args[0] += curr_pos[0]
    c_args[1] += curr_pos[1]
    c_args[2] += curr_pos[0]
    c_args[3] += curr_pos[1]
    c_args[4] += curr_pos[0]
    c_args[5] += curr_pos[1]
  x, y = _cubicbezier(curr_pos[0], curr_pos[1],
                      c_args[0], c_args[1],
                      c_args[2], c_args[3],
                      c_args[4], c_args[5])
  max_possible = len(canvas)
  x = [int(round(x_)) for x_ in x]
  y = [int(round(y_)) for y_ in y]
  within_range = lambda x: 0 <= x < max_possible
  filtered = [(x_, y_) for x_, y_ in zip(x, y)
              if within_range(x_) and within_range(y_)]
  if not filtered:
    return
  x, y = list(zip(*filtered))
  canvas[y, x, :] = color


def _render_line(canvas, curr_pos, l_args, absolute, color):
  """Renders a line in the given canvas."""
  end_point = l_args
  if not absolute:
    end_point[0] += curr_pos[0]
    end_point[1] += curr_pos[1]
  rr, cc, val = draw.line_aa(int(curr_pos[0]), int(curr_pos[1]),
                             int(end_point[0]), int(end_point[1]))

  max_possible = len(canvas)
  within_range = lambda x: 0 <= x < max_possible
  filtered = [(x, y, v) for x, y, v in zip(rr, cc, val)
              if within_range(x) and within_range(y)]
  if not filtered:
    return
  rr, cc, val = list(zip(*filtered))
  val = [(v * color) for v in val]
  canvas[cc, rr, :] = val


def per_step_render(path, absolute=False, color=constant_color):
  """Render the icon's edges, given its path."""
  to_canvas_size = lambda l: [float(f) * (64./24.) for f in l]

  canvas = np.zeros((64, 64, 3))
  curr_pos = (0.0, 0.0)
  for i, cmd in enumerate(path):
    if not cmd: continue
    if cmd[0] in 'mM':
      curr_pos = _update_pos(curr_pos, to_canvas_size(cmd[-2:]), absolute)
    elif cmd[0] in 'cC':
      _render_cubic(canvas, curr_pos, to_canvas_size(cmd[1:]), absolute,
                    color(i, 55))
      curr_pos = _update_pos(curr_pos, to_canvas_size(cmd[-2:]), absolute)
    elif cmd[0] in 'lL':
      _render_line(canvas, curr_pos, to_canvas_size(cmd[1:]), absolute,
                   color(i, 55))
      curr_pos = _update_pos(curr_pos, to_canvas_size(cmd[1:]), absolute)

  return canvas


def zoom_out(path_list, add_baseline=0., per=22):
  """Makes glyph slightly smaller in viewbox, makes some descenders visible."""
  # assumes tensor is already unnormalized, and in long form
  new_path = []
  for command in path_list:
    args = []
    is_even = False
    for arg in command[1:]:
      if is_even:
        args.append(str(float(arg) - ((24.-per) / 24.)*64./4.))
        is_even = False
      else:
        args.append(str(float(arg) - add_baseline))
        is_even = True
    new_path.append([command[0]] + args)
  return new_path


##################### UTILS FOR PROCESSING VECTORS ################
def append_eos(sample, categorical, feature_dim):
  if not categorical:
    eos = -1 * np.ones(feature_dim)
  else:
    eos = np.zeros(feature_dim)
    eos[0] = 1.0
  sample.append(eos)
  return sample


def make_simple_cmds_long(out):
  """Converts svg decoder output to format required by some render functions."""
  # out has 10 dims
  # the first 4 are respectively dims 0, 4, 5, 9 of the full 20-dim onehot vec
  # the latter 6 are the 6 last dims of the 10-dim arg vec
  shape_minus_dim = list(np.shape(out))[:-1]
  return np.concatenate([out[..., :1],
                         np.zeros(shape_minus_dim + [3]),
                         out[..., 1:3],
                         np.zeros(shape_minus_dim + [3]),
                         out[..., 3:4],
                         np.zeros(shape_minus_dim + [14]),
                         out[..., 4:]], -1)


################# UTILS FOR CONVERTING VECTORS TO SVGS ########################
def vector_to_svg(vectors, stop_at_eos=False, categorical=False):
  """Tranforms a given vector to an svg string."""
  new_path = []
  for vector in vectors:
    if stop_at_eos:
      if categorical:
        try:
          is_eos = np.argmax(vector[:len(CMDS_LIST) + 1]) == 0
        except:
          raise Exception(vector)
      else:
        is_eos = vector[0] < -0.5

      if is_eos:
        break
    new_path.append(' '.join(vector_to_cmd(vector, categorical=categorical)))
  new_path = ' '.join(new_path)
  return SVG_PREFIX_BIG + PATH_PREFIX_1 + new_path + PATH_POSFIX_1 + SVG_POSFIX


def vector_to_cmd(vector, categorical=False, return_floats=False):
  """Does the inverse transformation as cmd_to_vector()."""
  cast_fn = float if return_floats else str
  if categorical:
    command = (vector[:len(CMDS_LIST) + 1],)
    arguments = vector[len(CMDS_LIST) + 1:]
    cmd_idx = np.argmax(command) - 1
  else:
    command, arguments = vector[:1], vector[1:]
    cmd_idx = int(round(command[0]))

  if cmd_idx < -0.5:
    # EOS
    return []
  if cmd_idx >= len(CMDS_LIST):
    cmd_idx = len(CMDS_LIST) - 1

  cmd = CMDS_LIST[cmd_idx]
  cmd_list = [cmd]

  if cmd in 'hH':
    cmd_list.append(cast_fn(arguments[8]))  # x
  elif cmd in 'vV':
    cmd_list.append(cast_fn(arguments[9]))  # y
  elif cmd in 'mMlLtT':
    cmd_list.append(cast_fn(arguments[8]))  # x
    cmd_list.append(cast_fn(arguments[9]))  # y
  elif cmd in 'sSqQ':
    cmd_list.append(cast_fn(arguments[6]))  # x2
    cmd_list.append(cast_fn(arguments[7]))  # y2
    cmd_list.append(cast_fn(arguments[8]))  # x
    cmd_list.append(cast_fn(arguments[9]))  # y
  elif cmd in 'cC':
    cmd_list.append(cast_fn(arguments[4]))  # x1
    cmd_list.append(cast_fn(arguments[5]))  # y1
    cmd_list.append(cast_fn(arguments[6]))  # x2
    cmd_list.append(cast_fn(arguments[7]))  # y2
    cmd_list.append(cast_fn(arguments[8]))  # x
    cmd_list.append(cast_fn(arguments[9]))  # y
  elif cmd in 'aA':
    cmd_list.append(cast_fn(arguments[0]))  # rx
    cmd_list.append(cast_fn(arguments[1]))  # ry
    # x-axis-rotation is always 0
    cmd_list.append(cast_fn('0'))
    # the following two flags are binary.
    cmd_list.append(cast_fn(1 if arguments[2] > 0.5 else 0))  # large-arc-flag
    cmd_list.append(cast_fn(1 if arguments[3] > 0.5 else 0))  # sweep-flag
    cmd_list.append(cast_fn(arguments[8]))  # x
    cmd_list.append(cast_fn(arguments[9]))  # y

  return cmd_list


############## UTILS FOR CONVERTING SVGS/VECTORS TO IMAGES ###################
def create_image_conversion_fn(max_outputs, categorical=False):
  """Binds the number of outputs to the image conversion fn (to svg or png)."""
  def convert_to_svg(decoder_output):
    converted = []
    for example in decoder_output:
      if len(converted) == max_outputs:
        break
      converted.append(vector_to_svg(example, True, categorical=categorical))
    return np.array(converted)

  return convert_to_svg


################### UTILS FOR CREATING TF SUMMARIES ##########################
def _make_encoded_image(img_tensor):
  pil_img = Image.fromarray(np.squeeze(img_tensor * 255).astype(np.uint8),
                            mode='L')
  buff = io.BytesIO()
  pil_img.save(buff, format='png')
  encoded_image = buff.getvalue()
  return encoded_image


def make_text_summary_value(svg, tag):
  """Converts the given str to a text tf.summary.Summary.Value."""
  svg_proto = tf.make_tensor_proto(svg, tf.string)
  value = tf.summary.Summary.Value(tag=tag, tensor=svg_proto)
  value.metadata.plugin_data.plugin_name = 'text'
  return value


def make_image_summary(image_tensor, tag):
  """Converts the given image tensor to a tf.summary.Summary.Image."""
  encoded_image = _make_encoded_image(image_tensor)
  image_sum = tf.summary.Summary.Image(encoded_image_string=encoded_image)
  return tf.summary.Summary.Value(tag=tag, image=image_sum)



# Copyright 2020 The Magenta Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pipeline to create PolyphonyRNN dataset."""

from magenta.models.polyphony_rnn import polyphony_lib
from magenta.pipelines import dag_pipeline
from magenta.pipelines import event_sequence_pipeline
from magenta.pipelines import note_sequence_pipelines
from magenta.pipelines import pipeline
from magenta.pipelines import pipelines_common
from note_seq.protobuf import music_pb2


class PolyphonicSequenceExtractor(pipeline.Pipeline):
  """Extracts polyphonic tracks from a quantized NoteSequence."""

  def __init__(self, min_steps, max_steps, name=None):
    super(PolyphonicSequenceExtractor, self).__init__(
        input_type=music_pb2.NoteSequence,
        output_type=polyphony_lib.PolyphonicSequence,
        name=name)
    self._min_steps = min_steps
    self._max_steps = max_steps

  def transform(self, quantized_sequence):
    poly_seqs, stats = polyphony_lib.extract_polyphonic_sequences(
        quantized_sequence,
        min_steps_discard=self._min_steps,
        max_steps_discard=self._max_steps)
    self._set_stats(stats)
    return poly_seqs


def get_pipeline(config, min_steps, max_steps, eval_ratio):
  """Returns the Pipeline instance which creates the RNN dataset.

  Args:
    config: An EventSequenceRnnConfig.
    min_steps: Minimum number of steps for an extracted sequence.
    max_steps: Maximum number of steps for an extracted sequence.
    eval_ratio: Fraction of input to set aside for evaluation set.

  Returns:
    A pipeline.Pipeline instance.
  """
  # Transpose up to a major third in either direction.
  # Because our current dataset is Bach chorales, transposing more than a major
  # third in either direction probably doesn't makes sense (e.g., because it is
  # likely to exceed normal singing range).
  transposition_range = range(-4, 5)

  partitioner = pipelines_common.RandomPartition(
      music_pb2.NoteSequence,
      ['eval_poly_tracks', 'training_poly_tracks'],
      [eval_ratio])
  dag = {partitioner: dag_pipeline.DagInput(music_pb2.NoteSequence)}

  for mode in ['eval', 'training']:
    time_change_splitter = note_sequence_pipelines.TimeChangeSplitter(
        name='TimeChangeSplitter_' + mode)
    quantizer = note_sequence_pipelines.Quantizer(
        steps_per_quarter=config.steps_per_quarter, name='Quantizer_' + mode)
    transposition_pipeline = note_sequence_pipelines.TranspositionPipeline(
        transposition_range, name='TranspositionPipeline_' + mode)
    poly_extractor = PolyphonicSequenceExtractor(
        min_steps=min_steps, max_steps=max_steps, name='PolyExtractor_' + mode)
    encoder_pipeline = event_sequence_pipeline.EncoderPipeline(
        polyphony_lib.PolyphonicSequence, config.encoder_decoder,
        name='EncoderPipeline_' + mode)

    dag[time_change_splitter] = partitioner[mode + '_poly_tracks']
    dag[quantizer] = time_change_splitter
    dag[transposition_pipeline] = quantizer
    dag[poly_extractor] = transposition_pipeline
    dag[encoder_pipeline] = poly_extractor
    dag[dag_pipeline.DagOutput(mode + '_poly_tracks')] = encoder_pipeline

  return dag_pipeline.DAGPipeline(dag)



# Copyright 2020 The Magenta Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Beam search library."""

import collections
import copy
import heapq

# A beam entry containing a) the current sequence, b) a "state" containing any
# information needed to extend the sequence, and c) a score for the current
# sequence e.g. log-likelihood.
BeamEntry = collections.namedtuple('BeamEntry', ['sequence', 'state', 'score'])


def _generate_branches(beam_entries, generate_step_fn, branch_factor,
                       num_steps):
  """Performs a single iteration of branch generation for beam search.

  This method generates `branch_factor` branches for each sequence in the beam,
  where each branch extends the event sequence by `num_steps` steps (via calls
  to `generate_step_fn`). The resulting beam is returned.

  Args:
    beam_entries: A list of BeamEntry tuples, the current beam.
    generate_step_fn: A function that takes three parameters: a list of
        sequences, a list of states, and a list of scores, all of the same size.
        The function should generate a single step for each of the sequences and
        return the extended sequences, updated states, and updated (total)
        scores, as three lists. The function may modify the sequences, states,
        and scores in place, but should also return the modified values.
    branch_factor: The integer branch factor to use.
    num_steps: The integer number of steps to take per branch.

  Returns:
    The updated beam, with `branch_factor` times as many BeamEntry tuples.
  """
  if branch_factor > 1:
    branched_entries = beam_entries * branch_factor
    all_sequences = [copy.deepcopy(entry.sequence)
                     for entry in branched_entries]
    all_states = [copy.deepcopy(entry.state) for entry in branched_entries]
    all_scores = [entry.score for entry in branched_entries]
  else:
    # No need to make copies if there's no branching.
    all_sequences = [entry.sequence for entry in beam_entries]
    all_states = [entry.state for entry in beam_entries]
    all_scores = [entry.score for entry in beam_entries]

  for _ in range(num_steps):
    all_sequences, all_states, all_scores = generate_step_fn(
        all_sequences, all_states, all_scores)

  return [BeamEntry(sequence, state, score)
          for sequence, state, score
          in zip(all_sequences, all_states, all_scores)]


def _prune_branches(beam_entries, k):
  """Prune all but the `k` sequences with highest score from the beam."""
  indices = heapq.nlargest(k, range(len(beam_entries)),
                           key=lambda i: beam_entries[i].score)
  return [beam_entries[i] for i in indices]


def beam_search(initial_sequence, initial_state, generate_step_fn, num_steps,
                beam_size, branch_factor, steps_per_iteration):
  """Generates a sequence using beam search.

  Initially, the beam is filled with `beam_size` copies of the initial sequence.

  Each iteration, the beam is pruned to contain only the `beam_size` event
  sequences with highest score. Then `branch_factor` new event sequences are
  generated for each sequence in the beam. These new sequences are formed by
  extending each sequence in the beam by `steps_per_iteration` steps. So between
  a branching and a pruning phase, there will be `beam_size` * `branch_factor`
  active event sequences.

  After the final iteration, the single sequence in the beam with highest
  likelihood will be returned.

  The `generate_step_fn` function operates on lists of sequences + states +
  scores rather than single sequences. This is to allow for the possibility of
  batching.

  Args:
    initial_sequence: The initial sequence, a Python list-like object.
    initial_state: The state corresponding to the initial sequence, with any
        auxiliary information needed for extending the sequence.
    generate_step_fn: A function that takes three parameters: a list of
        sequences, a list of states, and a list of scores, all of the same size.
        The function should generate a single step for each of the sequences and
        return the extended sequences, updated states, and updated (total)
        scores, as three lists.
    num_steps: The integer length in steps of the final sequence, after
        generation.
    beam_size: The integer beam size to use.
    branch_factor: The integer branch factor to use.
    steps_per_iteration: The integer number of steps to take per iteration.

  Returns:
    A tuple containing a) the highest-scoring sequence as computed by the beam
    search, b) the state corresponding to this sequence, and c) the score of
    this sequence.
  """
  sequences = [copy.deepcopy(initial_sequence) for _ in range(beam_size)]
  states = [copy.deepcopy(initial_state) for _ in range(beam_size)]
  scores = [0] * beam_size

  beam_entries = [BeamEntry(sequence, state, score)
                  for sequence, state, score
                  in zip(sequences, states, scores)]

  # Choose the number of steps for the first iteration such that subsequent
  # iterations can all take the same number of steps.
  first_iteration_num_steps = (num_steps - 1) % steps_per_iteration + 1

  beam_entries = _generate_branches(
      beam_entries, generate_step_fn, branch_factor, first_iteration_num_steps)

  num_iterations = (num_steps -
                    first_iteration_num_steps) // steps_per_iteration

  for _ in range(num_iterations):
    beam_entries = _prune_branches(beam_entries, k=beam_size)
    beam_entries = _generate_branches(
        beam_entries, generate_step_fn, branch_factor, steps_per_iteration)

  # Prune to the single best beam entry.
  beam_entry = _prune_branches(beam_entries, k=1)[0]

  return beam_entry.sequence, beam_entry.state, beam_entry.score



# Copyright 2020 The Magenta Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Data processing pipelines for drum tracks."""

from magenta.pipelines import pipeline
from magenta.pipelines import statistics
from note_seq import drums_lib
from note_seq import DrumTrack
from note_seq import events_lib
from note_seq import sequences_lib
from note_seq.protobuf import music_pb2
import tensorflow.compat.v1 as tf


def extract_drum_tracks(quantized_sequence,
                        search_start_step=0,
                        min_bars=7,
                        max_steps_truncate=None,
                        max_steps_discard=None,
                        gap_bars=1.0,
                        pad_end=False,
                        ignore_is_drum=False):
  """Extracts a list of drum tracks from the given quantized NoteSequence.

  This function will search through `quantized_sequence` for drum tracks. A drum
  track can span multiple "tracks" in the sequence. Only one drum track can be
  active at a given time, but multiple drum tracks can be extracted from the
  sequence if gaps are present.

  Once a note-on drum event is encountered, a drum track begins. Gaps of silence
  will be splitting points that divide the sequence into separate drum tracks.
  The minimum size of these gaps are given in `gap_bars`. The size of a bar
  (measure) of music in time steps is computed form the time signature stored in
  `quantized_sequence`.

  A drum track is only used if it is at least `min_bars` bars long.

  After scanning the quantized NoteSequence, a list of all extracted DrumTrack
  objects is returned.

  Args:
    quantized_sequence: A quantized NoteSequence.
    search_start_step: Start searching for drums at this time step. Assumed to
        be the beginning of a bar.
    min_bars: Minimum length of drum tracks in number of bars. Shorter drum
        tracks are discarded.
    max_steps_truncate: Maximum number of steps in extracted drum tracks. If
        defined, longer drum tracks are truncated to this threshold. If pad_end
        is also True, drum tracks will be truncated to the end of the last bar
        below this threshold.
    max_steps_discard: Maximum number of steps in extracted drum tracks. If
        defined, longer drum tracks are discarded.
    gap_bars: A drum track comes to an end when this number of bars (measures)
        of no drums is encountered.
    pad_end: If True, the end of the drum track will be padded with empty events
        so that it will end at a bar boundary.
    ignore_is_drum: Whether accept notes where `is_drum` is False.

  Returns:
    drum_tracks: A python list of DrumTrack instances.
    stats: A dictionary mapping string names to `statistics.Statistic` objects.

  Raises:
    NonIntegerStepsPerBarError: If `quantized_sequence`'s bar length
        (derived from its time signature) is not an integer number of time
        steps.
  """
  drum_tracks = []
  stats = dict((stat_name, statistics.Counter(stat_name)) for stat_name in
               ['drum_tracks_discarded_too_short',
                'drum_tracks_discarded_too_long', 'drum_tracks_truncated'])
  # Create a histogram measuring drum track lengths (in bars not steps).
  # Capture drum tracks that are very small, in the range of the filter lower
  # bound `min_bars`, and large. The bucket intervals grow approximately
  # exponentially.
  stats['drum_track_lengths_in_bars'] = statistics.Histogram(
      'drum_track_lengths_in_bars',
      [0, 1, 10, 20, 30, 40, 50, 100, 200, 500, min_bars // 2, min_bars,
       min_bars + 1, min_bars - 1])

  steps_per_bar = int(
      sequences_lib.steps_per_bar_in_quantized_sequence(quantized_sequence))

  # Quantize the track into a DrumTrack object.
  # If any notes start at the same time, only one is kept.
  while 1:
    drum_track = DrumTrack()
    drum_track.from_quantized_sequence(
        quantized_sequence,
        search_start_step=search_start_step,
        gap_bars=gap_bars,
        pad_end=pad_end,
        ignore_is_drum=ignore_is_drum)
    search_start_step = (
        drum_track.end_step +
        (search_start_step - drum_track.end_step) % steps_per_bar)
    if not drum_track:
      break

    # Require a certain drum track length.
    if len(drum_track) < drum_track.steps_per_bar * min_bars:
      stats['drum_tracks_discarded_too_short'].increment()
      continue

    # Discard drum tracks that are too long.
    if max_steps_discard is not None and len(drum_track) > max_steps_discard:
      stats['drum_tracks_discarded_too_long'].increment()
      continue

    # Truncate drum tracks that are too long.
    if max_steps_truncate is not None and len(drum_track) > max_steps_truncate:
      truncated_length = max_steps_truncate
      if pad_end:
        truncated_length -= max_steps_truncate % drum_track.steps_per_bar
      drum_track.set_length(truncated_length)
      stats['drum_tracks_truncated'].increment()

    stats['drum_track_lengths_in_bars'].increment(
        len(drum_track) // drum_track.steps_per_bar)

    drum_tracks.append(drum_track)

  return drum_tracks, stats.values()


class DrumsExtractor(pipeline.Pipeline):
  """Extracts drum tracks from a quantized NoteSequence."""

  def __init__(self, min_bars=7, max_steps=512, gap_bars=1.0, name=None):
    super(DrumsExtractor, self).__init__(
        input_type=music_pb2.NoteSequence,
        output_type=drums_lib.DrumTrack,
        name=name)
    self._min_bars = min_bars
    self._max_steps = max_steps
    self._gap_bars = gap_bars

  def transform(self, quantized_sequence):
    try:
      # pylint has a false positive error on this method call for some reason.
      # pylint:disable=redundant-keyword-arg
      drum_tracks, stats = extract_drum_tracks(
          quantized_sequence,
          min_bars=self._min_bars,
          max_steps_truncate=self._max_steps,
          gap_bars=self._gap_bars)
      # pylint:enable=redundant-keyword-arg
    except events_lib.NonIntegerStepsPerBarError as detail:
      tf.logging.warning('Skipped sequence: %s', detail)
      drum_tracks = []
      stats = [statistics.Counter('non_integer_steps_per_bar', 1)]
    self._set_stats(stats)
    return drum_tracks



# Copyright 2020 The Magenta Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A library of functions that help with causal masking."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow.compat.v1 as tf


def shift_right(x):
  """Shift the input over by one and a zero to the front.

  Args:
    x: The [mb, time, channels] tensor input.

  Returns:
    x_sliced: The [mb, time, channels] tensor output.
  """
  shape = x.get_shape().as_list()
  x_padded = tf.pad(x, [[0, 0], [1, 0], [0, 0]])
  x_sliced = tf.slice(x_padded, [0, 0, 0], tf.stack([-1, shape[1], -1]))
  x_sliced.set_shape(shape)
  return x_sliced


def mul_or_none(a, b):
  """Return the element wise multiplicative of the inputs.

  If either input is None, we return None.

  Args:
    a: A tensor input.
    b: Another tensor input with the same type as a.

  Returns:
    None if either input is None. Otherwise returns a * b.
  """
  if a is None or b is None:
    return None
  return int(a * b)


def time_to_batch(x, block_size):
  """Splits time dimension (i.e. dimension 1) of `x` into batches.

  Within each batch element, the `k*block_size` time steps are transposed,
  so that the `k` time steps in each output batch element are offset by
  `block_size` from each other.

  The number of input time steps must be a multiple of `block_size`.

  Args:
    x: Tensor of shape [nb, k*block_size, n] for some natural number k.
    block_size: number of time steps (i.e. size of dimension 1) in the output
      tensor.

  Returns:
    Tensor of shape [nb*block_size, k, n]
  """
  shape = x.get_shape().as_list()
  y = tf.reshape(x, [
      shape[0], shape[1] // block_size, block_size, shape[2]
  ])
  y = tf.transpose(y, [0, 2, 1, 3])
  y = tf.reshape(y, [
      shape[0] * block_size, shape[1] // block_size, shape[2]
  ])
  y.set_shape([
      mul_or_none(shape[0], block_size), mul_or_none(shape[1], 1. / block_size),
      shape[2]
  ])
  return y


def batch_to_time(x, block_size):
  """Inverse of `time_to_batch(x, block_size)`.

  Args:
    x: Tensor of shape [nb*block_size, k, n] for some natural number k.
    block_size: number of time steps (i.e. size of dimension 1) in the output
      tensor.

  Returns:
    Tensor of shape [nb, k*block_size, n].
  """
  shape = x.get_shape().as_list()
  y = tf.reshape(x, [shape[0] // block_size, block_size, shape[1], shape[2]])
  y = tf.transpose(y, [0, 2, 1, 3])
  y = tf.reshape(y, [shape[0] // block_size, shape[1] * block_size, shape[2]])
  y.set_shape([mul_or_none(shape[0], 1. / block_size),
               mul_or_none(shape[1], block_size),
               shape[2]])
  return y


def conv1d(x,
           num_filters,
           filter_length,
           name,
           dilation=1,
           causal=True,
           kernel_initializer=tf.uniform_unit_scaling_initializer(1.0),
           biases_initializer=tf.constant_initializer(0.0),
           is_training=True):
  """Fast 1D convolution that supports causal padding and dilation.

  Args:
    x: The [mb, time, channels] float tensor that we convolve.
    num_filters: The number of filter maps in the convolution.
    filter_length: The integer length of the filter.
    name: The name of the scope for the variables.
    dilation: The amount of dilation.
    causal: Whether or not this is a causal convolution.
    kernel_initializer: The kernel initialization function.
    biases_initializer: The biases initialization function.
    is_training: Whether or not ot use traininable variables.

  Returns:
    y: The output of the 1D convolution.
  """
  batch_size, length, num_input_channels = x.get_shape().as_list()
  assert length % dilation == 0

  kernel_shape = [1, filter_length, num_input_channels, num_filters]
  strides = [1, 1, 1, 1]
  biases_shape = [num_filters]
  padding = 'VALID' if causal else 'SAME'

  with tf.variable_scope(name):
    weights = tf.get_variable(
        'W', shape=kernel_shape, initializer=kernel_initializer,
        trainable=is_training)
    biases = tf.get_variable(
        'biases', shape=biases_shape, initializer=biases_initializer,
        trainable=is_training)

  x_ttb = time_to_batch(x, dilation)
  if filter_length > 1 and causal:
    x_ttb = tf.pad(x_ttb, [[0, 0], [filter_length - 1, 0], [0, 0]])

  x_ttb_shape = x_ttb.get_shape().as_list()
  x_4d = tf.reshape(x_ttb, [x_ttb_shape[0], 1,
                            x_ttb_shape[1], num_input_channels])
  y = tf.nn.conv2d(x_4d, weights, strides, padding=padding)
  y = tf.nn.bias_add(y, biases)
  y_shape = y.get_shape().as_list()
  y = tf.reshape(y, [y_shape[0], y_shape[2], num_filters])
  y = batch_to_time(y, dilation)
  y.set_shape([batch_size, length, num_filters])
  return y


def pool1d(x, window_length, name, mode='avg', stride=None):
  """1D pooling function that supports multiple different modes.

  Args:
    x: The [mb, time, channels] float tensor that we are going to pool over.
    window_length: The amount of samples we pool over.
    name: The name of the scope for the variables.
    mode: The type of pooling, either avg or max.
    stride: The stride length.

  Returns:
    pooled: The [mb, time // stride, channels] float tensor result of pooling.
  """
  if mode == 'avg':
    pool_fn = tf.nn.avg_pool
  elif mode == 'max':
    pool_fn = tf.nn.max_pool

  stride = stride or window_length
  batch_size, length, num_channels = x.get_shape().as_list()
  assert length % window_length == 0
  assert length % stride == 0

  window_shape = [1, 1, window_length, 1]
  strides = [1, 1, stride, 1]
  x_4d = tf.reshape(x, [batch_size, 1, length, num_channels])
  pooled = pool_fn(x_4d, window_shape, strides, padding='SAME', name=name)
  return tf.reshape(pooled, [batch_size, length // stride, num_channels])



# Copyright 2020 The Magenta Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Helper functions for generating sounds.
"""

from magenta.models.gansynth.lib import util
import note_seq
import numpy as np
import scipy.io.wavfile as wavfile

MAX_NOTE_LENGTH = 3.0
MAX_VELOCITY = 127.0


def slerp(p0, p1, t):
  """Spherical linear interpolation."""
  omega = np.arccos(np.dot(
      np.squeeze(p0/np.linalg.norm(p0)), np.squeeze(p1/np.linalg.norm(p1))))
  so = np.sin(omega)
  return np.sin((1.0-t)*omega) / so * p0 + np.sin(t*omega)/so * p1


def load_midi(midi_path, min_pitch=36, max_pitch=84):
  """Load midi as a notesequence."""
  midi_path = util.expand_path(midi_path)
  ns = note_seq.midi_file_to_sequence_proto(midi_path)
  pitches = np.array([n.pitch for n in ns.notes])
  velocities = np.array([n.velocity for n in ns.notes])
  start_times = np.array([n.start_time for n in ns.notes])
  end_times = np.array([n.end_time for n in ns.notes])
  valid = np.logical_and(pitches >= min_pitch, pitches <= max_pitch)
  notes = {'pitches': pitches[valid],
           'velocities': velocities[valid],
           'start_times': start_times[valid],
           'end_times': end_times[valid]}
  return ns, notes


def get_random_instruments(model, total_time, secs_per_instrument=2.0):
  """Get random latent vectors evenly spaced in time."""
  n_instruments = int(total_time / secs_per_instrument)
  z_instruments = model.generate_z(n_instruments)
  t_instruments = np.linspace(-.0001, total_time, n_instruments)
  return z_instruments, t_instruments


def get_z_notes(start_times, z_instruments, t_instruments):
  """Get interpolated latent vectors for each note."""
  z_notes = []
  for t in start_times:
    idx = np.searchsorted(t_instruments, t, side='left') - 1

    # Handles out of bounds errors.
    if idx.item() == t_instruments.size - 1:
      idx -= 1

    t_left = t_instruments[idx]
    t_right = t_instruments[idx + 1]
    interp = (t - t_left) / (t_right - t_left)
    z_notes.append(slerp(z_instruments[idx], z_instruments[idx + 1], interp))
  z_notes = np.vstack(z_notes)
  return z_notes


def get_envelope(t_note_length, t_attack=0.010, t_release=0.3, sr=16000):
  """Create an attack sustain release amplitude envelope."""
  t_note_length = min(t_note_length, MAX_NOTE_LENGTH)
  i_attack = int(sr * t_attack)
  i_sustain = int(sr * t_note_length)
  i_release = int(sr * t_release)
  i_tot = i_sustain + i_release  # attack envelope doesn't add to sound length
  envelope = np.ones(i_tot)
  # Linear attack
  envelope[:i_attack] = np.linspace(0.0, 1.0, i_attack)
  # Linear release
  envelope[i_sustain:i_tot] = np.linspace(1.0, 0.0, i_release)
  return envelope


def combine_notes(audio_notes, start_times, end_times, velocities, sr=16000):
  """Combine audio from multiple notes into a single audio clip.

  Args:
    audio_notes: Array of audio [n_notes, audio_samples].
    start_times: Array of note starts in seconds [n_notes].
    end_times: Array of note ends in seconds [n_notes].
    velocities: Array of velocity values [n_notes].
    sr: Integer, sample rate.

  Returns:
    audio_clip: Array of combined audio clip [audio_samples]
  """
  n_notes = len(audio_notes)
  clip_length = end_times.max() + MAX_NOTE_LENGTH
  audio_clip = np.zeros(int(clip_length) * sr)

  for t_start, t_end, vel, i in zip(
      start_times, end_times, velocities, range(n_notes)):
    # Generate an amplitude envelope
    t_note_length = t_end - t_start
    envelope = get_envelope(t_note_length)
    length = len(envelope)
    audio_note = audio_notes[i, :length] * envelope
    # Normalize
    audio_note /= audio_note.max()
    audio_note *= (vel / MAX_VELOCITY)
    # Add to clip buffer
    clip_start = int(t_start * sr)
    clip_end = clip_start + length
    audio_clip[clip_start:clip_end] += audio_note

  # Normalize
  audio_clip /= audio_clip.max()
  audio_clip /= 2.0
  return audio_clip


def save_wav(audio, fname, sr=16000):
  wavfile.write(fname, sr, audio.astype('float32'))
  print('Saved to {}'.format(fname))



# Copyright 2020 The Magenta Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python2, python3
"""Transform one or multiple video in a set of frames.

Files are prefixed by a f followed by the frame number.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import argparse
import glob
import os
import sys

from PIL import Image
import six
import skvideo.io

PARSER = argparse.ArgumentParser(description="""
Transform one or multiple video in a set of frames.

Files are prefixed by a f followed by the frame number""")

PARSER.add_argument(
    '--video_in',
    dest='video_in',
    help="""one video or a path and a wildcard,
            wildcard need to be inside a quote,
            please note that ~ can be expanded only outside quote
            for instance ~/test.'*' works, but '~/test.*' won't""",
    required=True)
PARSER.add_argument(
    '--from',
    dest='from_s',
    type=float,
    default=-1,
    help='starting time in second (-1)')
PARSER.add_argument(
    '--to',
    dest='to_s',
    type=float,
    default=-1,
    help='last time in second (-1)')
PARSER.add_argument(
    '--path_out', dest='path_out', default='./', help='Destination folder (./)')
PARSER.add_argument(
    '--offset',
    dest='offset',
    type=int,
    default=0,
    help="""skip first frame to offset the output (0)
            useful with '--skip' to extract only a subset""")
PARSER.add_argument(
    '--skip',
    dest='skip',
    type=int,
    default=1,
    help='"--skip n" will extract every n frames (1)')
PARSER.add_argument(
    '--size',
    dest='size',
    type=int,
    default=256,
    help='size (256), this argument is used, only if cropped')
PARSER.add_argument(
    '--start',
    dest='start',
    type=int,
    default=0,
    help='starting number for the filename (0)')
PARSER.add_argument(
    '--multiple',
    dest='multiple',
    type=int,
    default=10000,
    help=
    '''if used with a wildcard (*),
    "multiple" will be added for each video (10000)'''
)
PARSER.add_argument(
    '--format', dest='format_ext', default='jpg', help='(jpg) or png')
PARSER.add_argument(
    '--crop',
    dest='crop',
    action='store_true',
    help='by default the video is cropped')
PARSER.add_argument(
    '--strech',
    dest='crop',
    action='store_false',
    help='the video can be streched to a square ratio')
PARSER.set_defaults(crop=True)

ARGS = PARSER.parse_args()


def crop(img, size):
  """resize the images.

  Args:
    img: a pillow image
    size: the size of the image (both x & y)

  Returns:
    nothing
  """
  small_side = min(img.size)
  center = img.size[0] / 2
  margin_left = center - small_side / 2
  margin_right = margin_left + small_side
  img = img.crop((margin_left, 0, margin_right, small_side))
  img = img.resize((size, size), Image.ANTIALIAS)
  return img


def main(_):
  """The main fonction use skvideo to extract frames as jpg.

  It can do it from a part or the totality of the video.

  Args:
    Nothing
  """
  print('argument to expand', ARGS.video_in)
  print('argument expanded', glob.glob(ARGS.video_in))
  video_count = 0
  for video_filename in glob.glob(ARGS.video_in):
    print('start parsing', video_filename)
    data = skvideo.io.ffprobe(video_filename)['video']
    rate_str = six.ensure_str(data['@r_frame_rate']).split('/')
    rate = float(rate_str[0]) / float(rate_str[1])
    print('detected frame rate:', rate)

    print('load frames:')
    video = skvideo.io.vreader(video_filename)
    frame_count = 0
    file_count = 0
    for frame in video:
      if (frame_count > ARGS.offset) and \
         ((frame_count-ARGS.offset)%ARGS.skip == 0) and \
         (frame_count/rate >= ARGS.from_s) and \
         (frame_count/rate <= ARGS.to_s or ARGS.to_s == -1):
        print(frame_count,)
        img = Image.fromarray(frame)
        if ARGS.crop:
          img = crop(img, ARGS.size)
        # save file
        file_number = file_count + video_count * ARGS.multiple + ARGS.start
        if ARGS.format_ext.lower() == 'jpg':
          file_out = os.path.join(ARGS.path_out,
                                  'f{:07d}.jpg'.format(file_number))
          img.save(file_out, 'JPEG')
        elif ARGS.format_ext.lower() == 'png':
          file_out = os.path.join(ARGS.path_out,
                                  'f{:07d}.png'.format(file_number))
          img.save(file_out, 'PNG')
        else:
          print('unrecognize format', ARGS.format_ext)
          sys.exit()
        file_count += 1
      frame_count += 1
    video_count += 1


if __name__ == '__main__':
  main(0)



# Copyright 2020 The Magenta Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Layers for a progressive GAN model.

This module contains basic building blocks to build a progressive GAN model.

See https://arxiv.org/abs/1710.10196 for details about the model.

See https://github.com/tkarras/progressive_growing_of_gans for the original
theano implementation.
"""

import numpy as np
import tensorflow.compat.v1 as tf


def pixel_norm(images, epsilon=1.0e-8):
  """Pixel normalization.

  For each pixel a[i,j,k] of image in HWC format, normalize its value to
  b[i,j,k] = a[i,j,k] / SQRT(SUM_k(a[i,j,k]^2) / C + eps).

  Args:
    images: A 4D `Tensor` of NHWC format.
    epsilon: A small positive number to avoid division by zero.

  Returns:
    A 4D `Tensor` with pixel-wise normalized channels.
  """
  return images * tf.rsqrt(
      tf.reduce_mean(tf.square(images), axis=3, keepdims=True) + epsilon)


def _get_validated_scale(scale):
  """Returns the scale guaranteed to be a positive integer."""
  scale = int(scale)
  if scale <= 0:
    raise ValueError('`scale` must be a positive integer.')
  return scale


def downscale(images, scale):
  """Box downscaling of images.

  Args:
    images: A 4D `Tensor` in NHWC format.
    scale: A positive integer scale.

  Returns:
    A 4D `Tensor` of `images` down scaled by a factor `scale`.

  Raises:
    ValueError: If `scale` is not a positive integer.
  """
  scale = _get_validated_scale(scale)
  if scale == 1:
    return images
  return tf.nn.avg_pool(
      images,
      ksize=[1, scale, scale, 1],
      strides=[1, scale, scale, 1],
      padding='VALID')


def upscale(images, scale):
  """Box upscaling (also called nearest neighbors) of images.

  Args:
    images: A 4D `Tensor` in NHWC format.
    scale: A positive integer scale.

  Returns:
    A 4D `Tensor` of `images` up scaled by a factor `scale`.

  Raises:
    ValueError: If `scale` is not a positive integer.
  """
  scale = _get_validated_scale(scale)
  if scale == 1:
    return images
  return tf.batch_to_space(
      tf.tile(images, [scale**2, 1, 1, 1]),
      crops=[[0, 0], [0, 0]],
      block_size=scale)


def downscale_height(images, scale):
  """Box downscaling images along the H (axis=1) dimension.

  Args:
    images: A 4D `Tensor` in NHWC format.
    scale: A positive integer scale.

  Returns:
    A 4D `Tensor` of `images` down scaled by a factor `scale`.

  Raises:
    ValueError: If `scale` is not a positive integer.
  """
  scale = _get_validated_scale(scale)
  if scale == 1:
    return images
  return tf.nn.avg_pool(
      images, ksize=[1, scale, 1, 1], strides=[1, scale, 1, 1], padding='VALID')


def upscale_height(images, scale):
  """Box upscaling along the H (axis=1) dimension.

  Args:
    images: A 4D `Tensor` in NHWC format.
    scale: A positive integer scale.

  Returns:
    A 4D `Tensor` of `images` up scaled by a factor `scale`.

  Raises:
    ValueError: If `scale` is not a positive integer.
  """
  scale = _get_validated_scale(scale)
  if scale == 1:
    return images
  images = tf.batch_to_space_nd(
      tf.tile(images, [scale, 1, 1, 1]), block_shape=[scale], crops=[[0, 0]])
  return images


def minibatch_mean_stddev(x):
  """Computes the standard deviation average.

  This is used by the discriminator as a form of batch discrimination.

  Args:
    x: A `Tensor` for which to compute the standard deviation average. The first
        dimension must be batch size.

  Returns:
    A scalar `Tensor` which is the mean variance of variable x.
  """
  mean, var = tf.nn.moments(x, axes=[0])
  del mean
  return tf.reduce_mean(tf.sqrt(var + 1e-6))


def scalar_concat(tensor, scalar):
  """Concatenates a scalar to the last dimension of a tensor.

  Args:
    tensor: A `Tensor`.
    scalar: a scalar `Tensor` to concatenate to tensor `tensor`.

  Returns:
    A `Tensor`. If `tensor` has shape [...,N], the result R has shape
    [...,N+1] and R[...,N] = scalar.

  Raises:
    ValueError: If `tensor` is a scalar `Tensor`.
  """
  ndims = tensor.shape.ndims
  if ndims < 1:
    raise ValueError('`tensor` must have number of dimensions >= 1.')
  shape = tf.shape(tensor)
  return tf.concat(
      [tensor,
       tf.ones([shape[i] for i in range(ndims - 1)] + [1]) * scalar],
      axis=ndims - 1)


def he_initializer_scale(shape, slope=1.0):
  """The scale of He neural network initializer.

  Args:
    shape: A list of ints representing the dimensions of a tensor.
    slope: A float representing the slope of the ReLu following the layer.

  Returns:
    A float of he initializer scale.
  """
  fan_in = np.prod(shape[:-1])
  return np.sqrt(2. / ((1. + slope**2) * fan_in))


def debugprint(x, name=''):
  """Small wrapper for tf.Print which prints summary statistics."""
  name += '\t' + x.name
  return tf.Print(x,
                  [tf.reduce_min(x), tf.reduce_mean(x), tf.reduce_max(x)],
                  name)


def _custom_layer_impl(apply_kernel, kernel_shape, bias_shape, activation,
                       he_initializer_slope, use_weight_scaling):
  """Helper function to implement custom_xxx layer.

  Args:
    apply_kernel: A function that transforms kernel to output.
    kernel_shape: An integer tuple or list of the kernel shape.
    bias_shape: An integer tuple or list of the bias shape.
    activation: An activation function to be applied. None means no
        activation.
    he_initializer_slope: A float slope for the He initializer.
    use_weight_scaling: Whether to apply weight scaling.

  Returns:
    A `Tensor` computed as apply_kernel(kernel) + bias where kernel is a
    `Tensor` variable with shape `kernel_shape`, bias is a `Tensor` variable
    with shape `bias_shape`.
  """
  kernel_scale = he_initializer_scale(kernel_shape, he_initializer_slope)
  init_scale, post_scale = kernel_scale, 1.0
  if use_weight_scaling:
    init_scale, post_scale = post_scale, init_scale

  kernel_initializer = tf.random_normal_initializer(stddev=init_scale)

  bias = tf.get_variable(
      'bias', shape=bias_shape, initializer=tf.zeros_initializer())

  output = post_scale * apply_kernel(kernel_shape, kernel_initializer) + bias

  if activation is not None:
    output = activation(output)
  return output


def custom_conv2d(x,
                  filters,
                  kernel_size,
                  strides=(1, 1),
                  padding='SAME',
                  activation=None,
                  he_initializer_slope=1.0,
                  use_weight_scaling=True,
                  scope='custom_conv2d',
                  reuse=None):
  """Custom conv2d layer.

  In comparison with tf.layers.conv2d this implementation use the He initializer
  to initialize convolutional kernel and the weight scaling trick (if
  `use_weight_scaling` is True) to equalize learning rates. See
  https://arxiv.org/abs/1710.10196 for more details.

  Args:
    x: A `Tensor` of NHWC format.
    filters: An int of output channels.
    kernel_size: An integer or a int tuple of [kernel_height, kernel_width].
    strides: A list of strides.
    padding: One of "VALID" or "SAME".
    activation: An activation function to be applied. None means no
        activation. Defaults to None.
    he_initializer_slope: A float slope for the He initializer. Defaults to 1.0.
    use_weight_scaling: Whether to apply weight scaling. Defaults to True.
    scope: A string or variable scope.
    reuse: Whether to reuse the weights. Defaults to None.

  Returns:
    A `Tensor` of NHWC format where the last dimension has size `filters`.
  """
  if not isinstance(kernel_size, (list, tuple)):
    kernel_size = [kernel_size] * 2
  kernel_size = list(kernel_size)

  def _apply_kernel(kernel_shape, kernel_initializer):
    return tf.layers.conv2d(
        x,
        filters=filters,
        kernel_size=kernel_shape[0:2],
        strides=strides,
        padding=padding,
        use_bias=False,
        kernel_initializer=kernel_initializer)

  with tf.variable_scope(scope, reuse=reuse):
    return _custom_layer_impl(
        _apply_kernel,
        kernel_shape=kernel_size + [x.shape.as_list()[3], filters],
        bias_shape=(filters,),
        activation=activation,
        he_initializer_slope=he_initializer_slope,
        use_weight_scaling=use_weight_scaling)


def custom_dense(x,
                 units,
                 activation=None,
                 he_initializer_slope=1.0,
                 use_weight_scaling=True,
                 scope='custom_dense',
                 reuse=None):
  """Custom dense layer.

  In comparison with tf.layers.dense This implementation use the He
  initializer to initialize weights and the weight scaling trick
  (if `use_weight_scaling` is True) to equalize learning rates. See
  https://arxiv.org/abs/1710.10196 for more details.

  Args:
    x: A `Tensor`.
    units: An int of the last dimension size of output.
    activation: An activation function to be applied. None means no
        activation. Defaults to None.
    he_initializer_slope: A float slope for the He initializer. Defaults to 1.0.
    use_weight_scaling: Whether to apply weight scaling. Defaults to True.
    scope: A string or variable scope.
    reuse: Whether to reuse the weights. Defaults to None.

  Returns:
    A `Tensor` where the last dimension has size `units`.
  """
  x = tf.layers.flatten(x)

  def _apply_kernel(kernel_shape, kernel_initializer):
    return tf.layers.dense(
        x,
        kernel_shape[1],
        use_bias=False,
        kernel_initializer=kernel_initializer)

  with tf.variable_scope(scope, reuse=reuse):
    return _custom_layer_impl(
        _apply_kernel,
        kernel_shape=(x.shape.as_list()[-1], units),
        bias_shape=(units,),
        activation=activation,
        he_initializer_slope=he_initializer_slope,
        use_weight_scaling=use_weight_scaling)



# Copyright 2020 The Magenta Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Flags utility object for handling hyperparameter state."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import base64
import json


class Flags(dict):
  """For storing and accessing flags."""
  __getattr__ = dict.get
  __setattr__ = dict.__setitem__
  __delattr__ = dict.__delitem__

  def print_values(self, indent=1):
    for k, v in sorted(self.items()):
      if isinstance(v, Flags):
        print('{}{}:'.format('\t'*indent, k))
        v.print_values(indent=indent+1)
      else:
        print('{}{}: {}'.format('\t'*indent, k, v))

  def load(self, other):
    def recursive_update(flags, source_dict):
      for k in source_dict:
        if isinstance(source_dict[k], dict):
          flags[k] = Flags()
          recursive_update(flags[k], source_dict[k])
        else:
          flags[k] = source_dict[k]
    recursive_update(self, other)

  def load_json(self, json_string):
    other = json.loads(json_string)
    self.load(other)

  def load_b64json(self, json_string):
    json_string = base64.b64decode(json_string)
    print('!!!!!!JSON STRING!!!!!!!!')
    print(json_string)
    self.load_json(json_string)

  def set_if_empty(self, key, val):
    if key not in self:
      self[key] = val



# Copyright 2020 The Magenta Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for note_sequence_pipelines."""

from absl.testing import absltest
from magenta.common import testing_lib as common_testing_lib
from magenta.pipelines import note_sequence_pipelines
from note_seq import sequences_lib
from note_seq import testing_lib
from note_seq.protobuf import music_pb2


class PipelineUnitsCommonTest(absltest.TestCase):

  def _unit_transform_test(self, unit, input_instance,
                           expected_outputs):
    outputs = unit.transform(input_instance)
    self.assertIsInstance(outputs, list)
    common_testing_lib.assert_set_equality(self, expected_outputs, outputs)
    self.assertEqual(unit.input_type, type(input_instance))
    if outputs:
      self.assertEqual(unit.output_type, type(outputs[0]))

  def testSplitter(self):
    note_sequence = common_testing_lib.parse_test_proto(
        music_pb2.NoteSequence,
        """
        time_signatures: {
          numerator: 4
          denominator: 4}
        tempos: {
          qpm: 60}""")
    testing_lib.add_track_to_sequence(
        note_sequence, 0,
        [(12, 100, 0.01, 10.0), (11, 55, 0.22, 0.50), (40, 45, 2.50, 3.50),
         (55, 120, 4.0, 4.01), (52, 99, 4.75, 5.0)])
    expected_sequences = sequences_lib.split_note_sequence(note_sequence, 1.0)

    unit = note_sequence_pipelines.Splitter(1.0)
    self._unit_transform_test(unit, note_sequence, expected_sequences)

  def testTimeChangeSplitter(self):
    note_sequence = common_testing_lib.parse_test_proto(
        music_pb2.NoteSequence,
        """
        time_signatures: {
          time: 2.0
          numerator: 3
          denominator: 4}
        tempos: {
          qpm: 60}""")
    testing_lib.add_track_to_sequence(
        note_sequence, 0,
        [(12, 100, 0.01, 10.0), (11, 55, 0.22, 0.50), (40, 45, 2.50, 3.50),
         (55, 120, 4.0, 4.01), (52, 99, 4.75, 5.0)])
    expected_sequences = sequences_lib.split_note_sequence_on_time_changes(
        note_sequence)

    unit = note_sequence_pipelines.TimeChangeSplitter()
    self._unit_transform_test(unit, note_sequence, expected_sequences)

  def testQuantizer(self):
    steps_per_quarter = 4
    note_sequence = common_testing_lib.parse_test_proto(
        music_pb2.NoteSequence,
        """
        time_signatures: {
          numerator: 4
          denominator: 4}
        tempos: {
          qpm: 60}""")
    testing_lib.add_track_to_sequence(
        note_sequence, 0,
        [(12, 100, 0.01, 10.0), (11, 55, 0.22, 0.50), (40, 45, 2.50, 3.50),
         (55, 120, 4.0, 4.01), (52, 99, 4.75, 5.0)])
    expected_quantized_sequence = sequences_lib.quantize_note_sequence(
        note_sequence, steps_per_quarter)

    unit = note_sequence_pipelines.Quantizer(steps_per_quarter)
    self._unit_transform_test(unit, note_sequence,
                              [expected_quantized_sequence])

  def testSustainPipeline(self):
    note_sequence = common_testing_lib.parse_test_proto(
        music_pb2.NoteSequence,
        """
        time_signatures: {
          numerator: 4
          denominator: 4}
        tempos: {
          qpm: 60}""")
    testing_lib.add_track_to_sequence(
        note_sequence, 0,
        [(11, 55, 0.22, 0.50), (40, 45, 2.50, 3.50), (55, 120, 4.0, 4.01)])
    testing_lib.add_control_changes_to_sequence(
        note_sequence, 0,
        [(0.0, 64, 127), (0.75, 64, 0), (2.0, 64, 127), (3.0, 64, 0),
         (3.75, 64, 127), (4.5, 64, 127), (4.8, 64, 0), (4.9, 64, 127),
         (6.0, 64, 0)])
    expected_sequence = sequences_lib.apply_sustain_control_changes(
        note_sequence)

    unit = note_sequence_pipelines.SustainPipeline()
    self._unit_transform_test(unit, note_sequence, [expected_sequence])

  def testStretchPipeline(self):
    note_sequence = common_testing_lib.parse_test_proto(
        music_pb2.NoteSequence,
        """
        time_signatures: {
          time: 1.0
          numerator: 4
          denominator: 4}
        tempos: {
          qpm: 60}""")
    testing_lib.add_track_to_sequence(
        note_sequence, 0,
        [(11, 55, 0.22, 0.50), (40, 45, 2.50, 3.50), (55, 120, 4.0, 4.01)])

    expected_sequences = [
        sequences_lib.stretch_note_sequence(note_sequence, 0.5),
        sequences_lib.stretch_note_sequence(note_sequence, 1.0),
        sequences_lib.stretch_note_sequence(note_sequence, 1.5)]

    unit = note_sequence_pipelines.StretchPipeline(
        stretch_factors=[0.5, 1.0, 1.5])
    self._unit_transform_test(unit, note_sequence, expected_sequences)

  def testTranspositionPipeline(self):
    note_sequence = common_testing_lib.parse_test_proto(
        music_pb2.NoteSequence,
        """
        time_signatures: {
          numerator: 4
          denominator: 4}
        tempos: {
          qpm: 60}""")
    tp = note_sequence_pipelines.TranspositionPipeline(range(0, 2))
    testing_lib.add_track_to_sequence(
        note_sequence, 0,
        [(12, 100, 1.0, 4.0)])
    testing_lib.add_track_to_sequence(
        note_sequence, 1,
        [(36, 100, 2.0, 2.01)],
        is_drum=True)
    transposed = tp.transform(note_sequence)
    self.assertLen(transposed, 2)
    self.assertLen(transposed[0].notes, 2)
    self.assertLen(transposed[1].notes, 2)
    self.assertEqual(12, transposed[0].notes[0].pitch)
    self.assertEqual(13, transposed[1].notes[0].pitch)
    self.assertEqual(36, transposed[0].notes[1].pitch)
    self.assertEqual(36, transposed[1].notes[1].pitch)

  def testTranspositionPipelineOutOfRangeNotes(self):
    note_sequence = common_testing_lib.parse_test_proto(
        music_pb2.NoteSequence,
        """
        time_signatures: {
          numerator: 4
          denominator: 4}
        tempos: {
          qpm: 60}""")
    tp = note_sequence_pipelines.TranspositionPipeline(
        range(-1, 2), min_pitch=0, max_pitch=12)
    testing_lib.add_track_to_sequence(
        note_sequence, 0,
        [(10, 100, 1.0, 2.0), (12, 100, 2.0, 4.0), (13, 100, 4.0, 5.0)])
    transposed = tp.transform(note_sequence)
    self.assertLen(transposed, 1)
    self.assertLen(transposed[0].notes, 3)
    self.assertEqual(9, transposed[0].notes[0].pitch)
    self.assertEqual(11, transposed[0].notes[1].pitch)
    self.assertEqual(12, transposed[0].notes[2].pitch)


if __name__ == '__main__':
  absltest.main()



# Copyright 2020 The Magenta Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""A MIDI interface to the sequence generators.

Captures monophonic input MIDI sequences and plays back responses from the
sequence generator.
"""
import functools
import re
import threading
import time

from magenta.interfaces.midi import midi_hub
from magenta.interfaces.midi import midi_interaction
from magenta.models.drums_rnn import drums_rnn_sequence_generator
from magenta.models.melody_rnn import melody_rnn_sequence_generator
from magenta.models.performance_rnn import performance_sequence_generator
from magenta.models.pianoroll_rnn_nade import pianoroll_rnn_nade_sequence_generator
from magenta.models.polyphony_rnn import polyphony_sequence_generator
from magenta.models.shared import sequence_generator_bundle
import six
import tensorflow.compat.v1 as tf

tf.disable_v2_behavior()

FLAGS = tf.app.flags.FLAGS

tf.app.flags.DEFINE_bool(
    'list_ports',
    False,
    'Only list available MIDI ports.')
tf.app.flags.DEFINE_string(
    'input_ports',
    'magenta_in',
    'Comma-separated list of names of the input MIDI ports.')
tf.app.flags.DEFINE_string(
    'output_ports',
    'magenta_out',
    'Comma-separated list of names of the output MIDI ports.')
tf.app.flags.DEFINE_bool(
    'passthrough',
    True,
    'Whether to pass input messages through to the output port.')
tf.app.flags.DEFINE_integer(
    'clock_control_number',
    None,
    'The control change number to use with value 127 as a signal for a tick of '
    'the external clock. If None, an internal clock is used that ticks once '
    'per bar based on the qpm.')
tf.app.flags.DEFINE_integer(
    'end_call_control_number',
    None,
    'The control change number to use with value 127 as a signal to end the '
    'call phrase on the next tick.')
tf.app.flags.DEFINE_integer(
    'panic_control_number',
    None,
    'The control change number to use with value 127 as a panic signal to '
    'close open notes and clear playback sequence.')
tf.app.flags.DEFINE_integer(
    'mutate_control_number',
    None,
    'The control change number to use with value 127 as a mutate signal to '
    'generate a new response using the current response sequence as a seed.')
tf.app.flags.DEFINE_integer(
    'min_listen_ticks_control_number',
    None,
    'The control change number to use for controlling minimum listen duration. '
    'The value for the control number will be used in clock ticks. Inputs less '
    'than this length will be ignored.')
tf.app.flags.DEFINE_integer(
    'max_listen_ticks_control_number',
    None,
    'The control change number to use for controlling maximum listen duration. '
    'The value for the control number will be used in clock ticks. After this '
    'number of ticks, a response will automatically be generated. A 0 value '
    'signifies infinite duration.')
tf.app.flags.DEFINE_integer(
    'response_ticks_control_number',
    None,
    'The control change number to use for controlling response duration. The '
    'value for the control number will be used in clock ticks. If not set, the '
    'response duration will match the call duration.')
tf.app.flags.DEFINE_integer(
    'temperature_control_number',
    None,
    'The control change number to use for controlling softmax temperature.'
    'The value of control changes with this number will be used to set the '
    'temperature in a linear range between 0.1 and 2.')
tf.app.flags.DEFINE_boolean(
    'allow_overlap',
    False,
    'Whether to allow the call to overlap with the response.')
tf.app.flags.DEFINE_boolean(
    'enable_metronome',
    True,
    'Whether to enable the metronome. Ignored if `clock_control_number` is '
    'provided.')
tf.app.flags.DEFINE_integer(
    'metronome_channel',
    1,
    'The 0-based MIDI channel to output the metronome on. Ignored if '
    '`enable_metronome` is False or `clock_control_number` is provided.')
tf.app.flags.DEFINE_integer(
    'qpm',
    120,
    'The quarters per minute to use for the metronome and generated sequence. '
    'Overriden by values of control change signals for `tempo_control_number`.')
tf.app.flags.DEFINE_integer(
    'tempo_control_number',
    None,
    'The control change number to use for controlling tempo. qpm will be set '
    'to 60 more than the value of the control change.')
tf.app.flags.DEFINE_integer(
    'loop_control_number',
    None,
    'The control number to use for determining whether to loop the response. '
    'A value of 127 turns looping on and any other value turns it off.')
tf.app.flags.DEFINE_string(
    'bundle_files',
    None,
    'A comma-separated list of the location of the bundle files to use.')
tf.app.flags.DEFINE_integer(
    'generator_select_control_number',
    None,
    'The control number to use for selecting between generators when multiple '
    'bundle files are specified. Required unless only a single bundle file is '
    'specified.')
tf.app.flags.DEFINE_integer(
    'state_control_number',
    None,
    'The control number to use for sending the state. A value of 0 represents '
    '`IDLE`, 1 is `LISTENING`, and 2 is `RESPONDING`.')
tf.app.flags.DEFINE_float(
    'playback_offset',
    0.0,
    'Time in seconds to adjust playback time by.')
tf.app.flags.DEFINE_integer(
    'playback_channel',
    0,
    'MIDI channel to send play events.')
tf.app.flags.DEFINE_boolean(
    'learn_controls',
    False,
    'Whether to allow programming of control flags on startup.')
tf.app.flags.DEFINE_string(
    'log', 'WARN',
    'The threshold for what messages will be logged. DEBUG, INFO, WARN, ERROR, '
    'or FATAL.')

_CONTROL_FLAGS = [
    'clock_control_number',
    'end_call_control_number',
    'panic_control_number',
    'mutate_control_number',
    'min_listen_ticks_control_number',
    'max_listen_ticks_control_number',
    'response_ticks_control_number',
    'temperature_control_number',
    'tempo_control_number',
    'loop_control_number',
    'generator_select_control_number',
    'state_control_number']

# A map from a string generator name to its class.
_GENERATOR_MAP = melody_rnn_sequence_generator.get_generator_map()
_GENERATOR_MAP.update(drums_rnn_sequence_generator.get_generator_map())
_GENERATOR_MAP.update(performance_sequence_generator.get_generator_map())
_GENERATOR_MAP.update(pianoroll_rnn_nade_sequence_generator.get_generator_map())
_GENERATOR_MAP.update(polyphony_sequence_generator.get_generator_map())


class CCMapper(object):
  """A class for mapping control change numbers to specific controls.

  Args:
    cc_map: A dictionary containing mappings from signal names to control
        change numbers (or None). This dictionary will be updated by the class.
    midi_hub_: An initialized MidiHub to receive inputs from.
  """

  def __init__(self, cc_map, midi_hub_):
    self._cc_map = cc_map
    self._signals = list(cc_map.keys())
    self._midi_hub = midi_hub_
    self._update_event = threading.Event()

  def _print_instructions(self):
    """Prints instructions for mapping control changes."""
    print('Enter the index of a signal to set the control change for, or `q` '
          'when done.')
    fmt = '{:>6}\t{:<20}\t{:>6}'
    print(fmt.format('Index', 'Control', 'Current'))
    for i, signal in enumerate(self._signals):
      print(fmt.format(i + 1, signal, self._cc_map.get(signal)))
    print('')

  def _update_signal(self, signal, msg):
    """Updates mapping for the signal to the message's control change.

    Args:
      signal: The name of the signal to update the control change for.
      msg: The mido.Message whose control change the signal should be set to.
    """
    if msg.control in list(self._cc_map.values()):
      print('Control number %d is already assigned. Ignoring.' % msg.control)
    else:
      self._cc_map[signal] = msg.control
      print('Assigned control number %d to `%s`.' % (msg.control, signal))
    self._update_event.set()

  def update_map(self):
    """Enters a loop that receives user input to set signal controls."""
    while True:
      print('')
      self._print_instructions()
      response = input('Selection: ')
      if response == 'q':
        return
      try:
        signal = self._signals[int(response) - 1]
      except (ValueError, IndexError):
        print('Invalid response:', response)
        continue
      self._update_event.clear()
      self._midi_hub.register_callback(
          functools.partial(self._update_signal, signal),
          midi_hub.MidiSignal(type='control_change'))
      print('Send a control signal using the control number you wish to '
            'associate with `%s`.' % signal)
      self._update_event.wait()


def _validate_flags():
  """Returns True if flag values are valid or prints error and returns False."""
  if FLAGS.list_ports:
    print("Input ports: '%s'" % (
        "', '".join(midi_hub.get_available_input_ports())))
    print("Ouput ports: '%s'" % (
        "', '".join(midi_hub.get_available_output_ports())))
    return False

  if FLAGS.bundle_files is None:
    print('--bundle_files must be specified.')
    return False

  if (len(FLAGS.bundle_files.split(',')) > 1 and
      FLAGS.generator_select_control_number is None):
    tf.logging.warning(
        'You have specified multiple bundle files (generators), without '
        'setting `--generator_select_control_number`. You will only be able to '
        'use the first generator (%s).',
        FLAGS.bundle_files[0])

  return True


def _load_generator_from_bundle_file(bundle_file):
  """Returns initialized generator from bundle file path or None if fails."""
  try:
    bundle = sequence_generator_bundle.read_bundle_file(bundle_file)
  except sequence_generator_bundle.GeneratorBundleParseError:
    print('Failed to parse bundle file: %s' % FLAGS.bundle_file)
    return None

  generator_id = bundle.generator_details.id
  if generator_id not in _GENERATOR_MAP:
    print("Unrecognized SequenceGenerator ID '%s' in bundle file: %s" % (
        generator_id, FLAGS.bundle_file))
    return None

  generator = _GENERATOR_MAP[generator_id](checkpoint=None, bundle=bundle)
  generator.initialize()
  print("Loaded '%s' generator bundle from file '%s'." % (
      bundle.generator_details.id, bundle_file))
  return generator


def _print_instructions():
  """Prints instructions for interaction based on the flag values."""
  print('')
  print('Instructions:')
  print('Start playing  when you want to begin the call phrase.')
  if FLAGS.end_call_control_number is not None:
    print('When you want to end the call phrase, signal control number %d '
          'with value 127, or stop playing and wait one clock tick.'
          % FLAGS.end_call_control_number)
  else:
    print('When you want to end the call phrase, stop playing and wait one '
          'clock tick.')
  print('Once the response completes, the interface will wait for you to '
        'begin playing again to start a new call phrase.')
  print('')
  print('To end the interaction, press CTRL-C.')


def main(unused_argv):
  tf.logging.set_verbosity(FLAGS.log)

  if not _validate_flags():
    return

  # Load generators.
  generators = []
  for bundle_file in FLAGS.bundle_files.split(','):
    generators.append(_load_generator_from_bundle_file(bundle_file))
    if generators[-1] is None:
      return

  # Initialize MidiHub.
  hub = midi_hub.MidiHub(FLAGS.input_ports.split(','),
                         FLAGS.output_ports.split(','),
                         midi_hub.TextureType.POLYPHONIC,
                         passthrough=FLAGS.passthrough,
                         playback_channel=FLAGS.playback_channel,
                         playback_offset=FLAGS.playback_offset)

  control_map = {
      re.sub('_control_number$', '', six.ensure_str(f)): FLAGS.__getattr__(f)
      for f in _CONTROL_FLAGS
  }
  if FLAGS.learn_controls:
    CCMapper(control_map, hub).update_map()

  if control_map['clock'] is None:
    # Set the tick duration to be a single bar, assuming a 4/4 time signature.
    clock_signal = None
    tick_duration = 4 * (60. / FLAGS.qpm)
  else:
    clock_signal = midi_hub.MidiSignal(
        control=control_map['clock'], value=127)
    tick_duration = None

  def _signal_from_control_map(name):
    if control_map[name] is None:
      return None
    return midi_hub.MidiSignal(control=control_map[name], value=127)

  end_call_signal = _signal_from_control_map('end_call')
  panic_signal = _signal_from_control_map('panic')
  mutate_signal = _signal_from_control_map('mutate')

  metronome_channel = (
      FLAGS.metronome_channel if FLAGS.enable_metronome else None)
  interaction = midi_interaction.CallAndResponseMidiInteraction(
      hub,
      generators,
      FLAGS.qpm,
      FLAGS.generator_select_control_number,
      clock_signal=clock_signal,
      tick_duration=tick_duration,
      end_call_signal=end_call_signal,
      panic_signal=panic_signal,
      mutate_signal=mutate_signal,
      allow_overlap=FLAGS.allow_overlap,
      metronome_channel=metronome_channel,
      min_listen_ticks_control_number=control_map['min_listen_ticks'],
      max_listen_ticks_control_number=control_map['max_listen_ticks'],
      response_ticks_control_number=control_map['response_ticks'],
      tempo_control_number=control_map['tempo'],
      temperature_control_number=control_map['temperature'],
      loop_control_number=control_map['loop'],
      state_control_number=control_map['state'])

  _print_instructions()

  interaction.start()
  try:
    while True:
      time.sleep(1)
  except KeyboardInterrupt:
    interaction.stop()

  print('Interaction stopped.')


def console_entry_point():
  tf.app.run(main)


if __name__ == '__main__':
  console_entry_point()



# Copyright 2020 The Magenta Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module contains a registry of dataset classes."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
from magenta.models.gansynth.lib import spectral_ops
import numpy as np
import tensorflow.compat.v1 as tf
import tensorflow_datasets as tfds

Counter = collections.Counter


class BaseDataset(object):
  """A base class for reading data from disk."""

  def __init__(self, config):
    self._config = config

  def provide_one_hot_labels(self, batch_size):
    """Provides one-hot labels."""
    raise NotImplementedError

  def provide_dataset(self):
    """Provides audio dataset."""
    raise NotImplementedError

  def get_pitch_counts(self):
    """Returns a dictionary {pitch value (int): count (int)}."""
    raise NotImplementedError

  def get_pitches(self, num_samples):
    """Returns pitch_counter for num_samples for given dataset."""
    all_pitches = []
    pitch_counts = self.get_pitch_counts()
    for k, v in pitch_counts.items():
      all_pitches.extend([k]*v)
    sample_pitches = np.random.choice(all_pitches, num_samples)
    pitch_counter = Counter(sample_pitches)
    return pitch_counter


class NSynthTfdsDataset(BaseDataset):
  """A dataset for reading NSynth from Tensorflow Datasets (TFDS)."""

  def _get_dataset_from_tfds(self):
    """Loads GANsynth subset of NSynth dataset from TFDS."""
    try:
      dataset = tfds.load(
          'nsynth/gansynth_subset:2.3.*',
          data_dir=self._config['tfds_data_dir'],
          split=tfds.Split.TRAIN,
          download=False)
    except AssertionError as e:
      tf.logging.warning(
          'To train with the NSynth dataset, you must either set the '
          '\tfds_data_dir\' hparam to \'gs://tfds-data/datasets\' (recommended '
          'if running on GCP) or to a local directory where the dataset has '
          'been downloaded. You can download the dataset with the command '
          '`python -m tensorflow_datasets.scripts.download_and_prepare '
          '--datasets=nsynth/gansynth_subset --data_dir=/path/to/local/dir`.')
      raise e
    dataset = dataset.shuffle(buffer_size=1000)
    dataset = dataset.repeat()
    return dataset

  def provide_one_hot_labels(self, batch_size):
    """Provides one hot labels."""
    pitch_counts = self.get_pitch_counts()
    pitches = sorted(pitch_counts.keys())
    counts = [pitch_counts[p] for p in pitches]
    indices = tf.reshape(
        tf.multinomial(tf.log([tf.to_float(counts)]), batch_size), [batch_size])
    one_hot_labels = tf.one_hot(indices, depth=len(pitches))
    return one_hot_labels

  def provide_dataset(self):
    """Provides dataset (audio, labels) of nsynth."""
    length = 64000
    channels = 1

    pitch_counts = self.get_pitch_counts()
    pitches = sorted(pitch_counts.keys())
    label_index_table = tf.lookup.StaticVocabularyTable(
        tf.lookup.KeyValueTensorInitializer(
            keys=pitches,
            values=np.arange(len(pitches)),
            key_dtype=tf.int64,
            value_dtype=tf.int64),
        num_oov_buckets=1)

    def _parse_nsynth(example):
      """Parsing function for NSynth dataset."""
      wave, label = example['audio'], example['pitch']
      wave = spectral_ops.crop_or_pad(wave[tf.newaxis, :, tf.newaxis],
                                      length,
                                      channels)[0]
      one_hot_label = tf.one_hot(
          label_index_table.lookup(label), depth=len(pitches))
      return wave, one_hot_label

    dataset = self._get_dataset_from_tfds()
    dataset = dataset.map(
        _parse_nsynth, num_parallel_calls=tf.data.experimental.AUTOTUNE)
    return dataset

  def get_pitch_counts(self):
    pitch_counts = {
        24: 551,
        25: 545,
        26: 561,
        27: 550,
        28: 549,
        29: 578,
        30: 547,
        31: 628,
        32: 626,
        33: 630,
        34: 670,
        35: 668,
        36: 751,
        37: 768,
        38: 788,
        39: 801,
        40: 821,
        41: 871,
        42: 873,
        43: 860,
        44: 873,
        45: 844,
        46: 883,
        47: 865,
        48: 1058,
        49: 1036,
        50: 1094,
        51: 1065,
        52: 1186,
        53: 1200,
        54: 1187,
        55: 1346,
        56: 1273,
        57: 1320,
        58: 1263,
        59: 1283,
        60: 1349,
        61: 1285,
        62: 1281,
        63: 1332,
        64: 1286,
        65: 1301,
        66: 1220,
        67: 1300,
        68: 1209,
        69: 1257,
        70: 1288,
        71: 1223,
        72: 1258,
        73: 1183,
        74: 1178,
        75: 1125,
        76: 1171,
        77: 1089,
        78: 1024,
        79: 1081,
        80: 966,
        81: 1043,
        82: 991,
        83: 959,
        84: 977,
    }
    return pitch_counts


registry = {
    'nsynth_tfds': NSynthTfdsDataset,
}



# Copyright 2020 The Magenta Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Utilities for training."""

import collections
import copy
import functools
import random
import sys
import tensorflow.compat.v1 as tf
import tf_slim


# Should not be called from within the graph to avoid redundant summaries.
def _trial_summary(hparams, model_dir, output_dir, additional_trial_info):
  """Writes a tensorboard text summary of the trial."""

  summaries_to_write = collections.OrderedDict()
  summaries_to_write['model_dir'] = model_dir
  summaries_to_write['command_line_args'] = ' \\'.join(sys.argv)

  tf.logging.info('Writing hparams summary: %s', hparams)

  hparams_dict = hparams.values()

  # Create a markdown table from hparams.
  header = '| Key | Value |\n| :--- | :--- |\n'
  keys = sorted(hparams_dict.keys())
  lines = ['| %s | %s |' % (key, str(hparams_dict[key])) for key in keys]
  hparams_table = header + '\n'.join(lines) + '\n'

  summaries_to_write['hparams'] = hparams_table

  summaries_to_write.update(additional_trial_info)

  with tf.Session() as sess:
    writer = tf.summary.FileWriter(output_dir, graph=sess.graph)
    for name, summary in summaries_to_write.items():
      tf.logging.info('Writing summary for %s: %s', name, summary)
      writer.add_summary(
          tf.summary.text(name, tf.constant(summary, name=name),
                          collections=[]).eval())
    writer.close()


def create_estimator(model_fn,
                     model_dir,
                     hparams,
                     use_tpu=False,
                     master='',
                     tpu_cluster=None,
                     save_checkpoint_steps=300,
                     save_summary_steps=300,
                     keep_checkpoint_max=None,
                     warm_start_from=None):
  """Creates an estimator."""
  def wrapped_model_fn(features, labels, mode, params, config):
    """Wrap model_fn to restore labels value if present in features."""
    # Workaround for Estimator API that forces 'labels' to be None when in
    # predict mode.
    # https://github.com/tensorflow/tensorflow/issues/17824
    # See also infer_util.labels_to_features_wrapper
    if labels is None and hasattr(features, 'labels'):
      labels = features.labels
    return model_fn(features, labels, mode, params, config)

  if tpu_cluster:
    tpu_cluster_resolver = tf.distribute.cluster_resolver.TPUClusterResolver(
        tpu_cluster)
    master = None
  else:
    tpu_cluster_resolver = None

  config = tf.estimator.tpu.RunConfig(
      tpu_config=tf.estimator.tpu.TPUConfig(
          iterations_per_loop=save_checkpoint_steps),
      master=master,
      cluster=tpu_cluster_resolver,
      save_summary_steps=save_summary_steps,
      save_checkpoints_steps=save_checkpoint_steps,
      keep_checkpoint_max=keep_checkpoint_max,
      keep_checkpoint_every_n_hours=1)

  params = copy.deepcopy(hparams)
  params.del_hparam('batch_size')
  return tf.estimator.tpu.TPUEstimator(
      use_tpu=use_tpu,
      model_fn=wrapped_model_fn,
      model_dir=model_dir,
      params=params,
      train_batch_size=hparams.batch_size,
      eval_batch_size=hparams.eval_batch_size,
      predict_batch_size=hparams.predict_batch_size,
      config=config,
      warm_start_from=warm_start_from,
      eval_on_tpu=False)


def train(master,
          tpu_cluster,
          model_fn,
          data_fn,
          additional_trial_info,
          model_dir,
          preprocess_examples,
          hparams,
          keep_checkpoint_max,
          use_tpu,
          num_steps=None):
  """Train loop."""
  estimator = create_estimator(
      model_fn=model_fn,
      model_dir=model_dir,
      master=master,
      tpu_cluster=tpu_cluster,
      hparams=hparams,
      keep_checkpoint_max=keep_checkpoint_max,
      use_tpu=use_tpu)

  if estimator.config.is_chief:
    _trial_summary(
        hparams=hparams,
        model_dir=model_dir,
        output_dir=model_dir,
        additional_trial_info=additional_trial_info)

  transcription_data = functools.partial(
      data_fn,
      preprocess_examples=preprocess_examples,
      is_training=True,
      shuffle_examples=True,
      skip_n_initial_records=0)

  estimator.train(input_fn=transcription_data, max_steps=num_steps)


def evaluate(master,
             model_fn,
             data_fn,
             additional_trial_info,
             model_dir,
             preprocess_examples,
             hparams,
             name,
             num_steps=None):
  """Evaluation loop."""
  estimator = create_estimator(
      model_fn=model_fn, model_dir=model_dir, master=master, hparams=hparams)

  transcription_data_base = functools.partial(
      data_fn,
      preprocess_examples=preprocess_examples,
      is_training=False)

  if num_steps is None:
    transcription_data = functools.partial(
        transcription_data_base,
        shuffle_examples=False, skip_n_initial_records=0)
  else:
    # If num_steps is specified, we will evaluate only a subset of the data.
    #
    # The following is a hack that works around the problems of not being able
    # to determine the number of records in a given TFRecord shard without
    # reading the whole thing and not being able to persist a tf.data.Dataset
    # session across multiple estimator evaluate calls.
    #
    # This code tries to select a different subset for every evaluation by doing
    # the following:
    # - Setting shuffle_examples=True. This shuffles not only individual
    #   examples, but also shuffles the order in which shards are read.
    # - Skipping N examples before starting evaluation, where N is selected
    #   randomly for each evaluation run. This provides a different starting
    #   offset.

    # In order to skip a random number of records, we need to provide an upper
    # bound that will still let us run num_steps evaluation steps before running
    # out of data. The following code does a one-time check on startup to see
    # if there are up to num_steps * 5 records available, which would allow
    # a maximum skip range of [0, num_steps*4].
    records_to_check = num_steps * 5
    tf.logging.info('Checking for at least %d records...', records_to_check)
    records_available = 0
    with tf.Graph().as_default():
      record_check_params = copy.deepcopy(hparams)
      record_check_params.batch_size = 1
      iterator = transcription_data_base(
          params=record_check_params,
          shuffle_examples=False,
          skip_n_initial_records=0,
          ).make_initializable_iterator()
      next_record = iterator.get_next()
      with tf.Session() as sess:
        sess.run(iterator.initializer)
        try:
          for i in range(records_to_check):
            del i
            sess.run(next_record)
            records_available += 1
            if records_available % 10 == 0:
              tf.logging.info('Found %d records...', records_available)
        except tf.errors.OutOfRangeError:
          pass
    # Determine max number of records we could skip and still have num_steps
    # records remaining.
    max_records_to_skip = max(0, records_available - num_steps)
    tf.logging.info('Found at least %d records. '
                    'Will skip a maximum of %d records during eval runs '
                    'in order to support %d evaluation steps.',
                    records_available, max_records_to_skip, num_steps)

    # Since we're doing a limited number of steps, we should shuffle the
    # examples we're evaluating so each evaluation is over a different portion
    # of the dataset.
    def transcription_data(params, *args, **kwargs):
      assert not args
      skip_n_initial_records = random.randint(0, max_records_to_skip)
      tf.logging.info('Skipping %d initial record(s)', skip_n_initial_records)
      return transcription_data_base(
          params=params,
          shuffle_examples=True,
          skip_n_initial_records=skip_n_initial_records,
          **kwargs)

  _trial_summary(
      hparams=hparams,
      model_dir=model_dir,
      output_dir=estimator.eval_dir(name),
      additional_trial_info=additional_trial_info)

  checkpoint_path = None
  while True:
    checkpoint_path = tf_slim.evaluation.wait_for_new_checkpoint(
        model_dir, last_checkpoint=checkpoint_path)
    estimator.evaluate(input_fn=transcription_data, steps=num_steps,
                       checkpoint_path=checkpoint_path, name=name)
